import torch

import torch_geometric.typing
from torch_geometric import EdgeIndex
from torch_geometric.nn import GraphConv
from torch_geometric.testing import is_full_test
from torch_geometric.typing import SparseTensor
from torch_geometric.utils import to_torch_csc_tensor


def test_graph_conv():
    x1 = torch.randn(4, 8)
    x2 = torch.randn(2, 16)
    edge_index = torch.tensor([[0, 1, 2, 3], [0, 0, 1, 1]])
    value = torch.randn(edge_index.size(1))
    adj1 = to_torch_csc_tensor(edge_index, size=(4, 4))
    adj2 = to_torch_csc_tensor(edge_index, value, size=(4, 4))

    conv = GraphConv(8, 32)
    assert str(conv) == 'GraphConv(8, 32)'
    out1 = conv(x1, edge_index)
    assert out1.size() == (4, 32)
    assert torch.allclose(conv(x1, edge_index, size=(4, 4)), out1, atol=1e-6)
    assert torch.allclose(conv(x1, adj1.t()), out1, atol=1e-6)

    assert conv(
        x1,
        EdgeIndex(edge_index, sort_order='col', sparse_size=(4, 4)),
    ).allclose(out1, atol=1e-6)

    if torch_geometric.typing.WITH_TORCH_SPARSE:
        adj3 = SparseTensor.from_edge_index(edge_index, sparse_sizes=(4, 4))
        assert torch.allclose(conv(x1, adj3.t()), out1, atol=1e-6)

    out2 = conv(x1, edge_index, value)
    assert out2.size() == (4, 32)
    assert torch.allclose(conv(x1, edge_index, value, size=(4, 4)), out2,
                          atol=1e-6)
    assert torch.allclose(conv(x1, adj2.t()), out2, atol=1e-6)

    assert conv(
        x1,
        EdgeIndex(edge_index, sort_order='col', sparse_size=(4, 4)),
        value,
    ).allclose(out2, atol=1e-6)

    if torch_geometric.typing.WITH_TORCH_SPARSE:
        adj4 = SparseTensor.from_edge_index(edge_index, value, (4, 4))
        assert torch.allclose(conv(x1, adj4.t()), out2, atol=1e-6)

    if is_full_test():
        jit = torch.jit.script(conv)
        assert torch.allclose(jit(x1, edge_index), out1)
        assert torch.allclose(jit(x1, edge_index, size=(4, 4)), out1)
        assert torch.allclose(jit(x1, edge_index, value), out2)
        assert torch.allclose(jit(x1, edge_index, value, size=(4, 4)), out2)

        if torch_geometric.typing.WITH_TORCH_SPARSE:
            assert torch.allclose(jit(x1, adj3.t()), out1, atol=1e-6)
            assert torch.allclose(jit(x1, adj4.t()), out2, atol=1e-6)

    # Test bipartite message passing:
    adj1 = to_torch_csc_tensor(edge_index, size=(4, 2))
    adj2 = to_torch_csc_tensor(edge_index, value, size=(4, 2))

    conv = GraphConv((8, 16), 32)
    assert str(conv) == 'GraphConv((8, 16), 32)'
    out1 = conv((x1, x2), edge_index)
    assert out1.size() == (2, 32)
    assert torch.allclose(conv((x1, x2), edge_index, size=(4, 2)), out1)
    assert torch.allclose(conv((x1, x2), adj1.t()), out1, atol=1e-6)

    assert conv(
        (x1, x2),
        EdgeIndex(edge_index, sort_order='col', sparse_size=(4, 2)),
    ).allclose(out1, atol=1e-6)

    out2 = conv((x1, None), edge_index, size=(4, 2))
    assert out2.size() == (2, 32)
    assert torch.allclose(conv((x1, None), adj1.t()), out2, atol=1e-6)

    assert conv(
        (x1, None),
        EdgeIndex(edge_index, sort_order='col', sparse_size=(4, 2)),
    ).allclose(out2, atol=1e-6)

    out3 = conv((x1, x2), edge_index, value)
    assert out3.size() == (2, 32)
    assert torch.allclose(conv((x1, x2), edge_index, value, (4, 2)), out3)
    assert torch.allclose(conv((x1, x2), adj2.t()), out3, atol=1e-6)

    assert conv(
        (x1, x2),
        EdgeIndex(edge_index, sort_order='col', sparse_size=(4, 2)),
        value,
    ).allclose(out3, atol=1e-6)

    out4 = conv((x1, None), edge_index, value, size=(4, 2))
    assert out4.size() == (2, 32)
    assert torch.allclose(conv((x1, None), adj2.t()), out4, atol=1e-6)

    assert conv(
        (x1, None),
        EdgeIndex(edge_index, sort_order='col', sparse_size=(4, 2)),
        value,
    ).allclose(out4, atol=1e-6)

    if torch_geometric.typing.WITH_TORCH_SPARSE:
        adj3 = SparseTensor.from_edge_index(edge_index, sparse_sizes=(4, 2))
        adj4 = SparseTensor.from_edge_index(edge_index, value, (4, 2))
        assert torch.allclose(conv((x1, x2), adj3.t()), out1, atol=1e-6)
        assert torch.allclose(conv((x1, None), adj3.t()), out2, atol=1e-6)
        assert torch.allclose(conv((x1, x2), adj3.t()), out1, atol=1e-6)
        assert torch.allclose(conv((x1, None), adj4.t()), out4, atol=1e-6)

    if is_full_test():
        jit = torch.jit.script(conv)
        assert torch.allclose(jit((x1, x2), edge_index), out1)
        assert torch.allclose(jit((x1, x2), edge_index, size=(4, 2)), out1)
        assert torch.allclose(jit((x1, None), edge_index, size=(4, 2)), out2)
        assert torch.allclose(jit((x1, x2), edge_index, value), out3)
        assert torch.allclose(jit((x1, x2), edge_index, value, (4, 2)), out3)
        assert torch.allclose(jit((x1, None), edge_index, value, (4, 2)), out4)

        if torch_geometric.typing.WITH_TORCH_SPARSE:
            assert torch.allclose(jit((x1, x2), adj3.t()), out1, atol=1e-6)
            assert torch.allclose(jit((x1, None), adj3.t()), out2, atol=1e-6)
            assert torch.allclose(jit((x1, x2), adj4.t()), out3, atol=1e-6)
            assert torch.allclose(jit((x1, None), adj4.t()), out4, atol=1e-6)


class EdgeGraphConv(GraphConv):
    def message(self, x_j, edge_weight):
        return edge_weight.view(-1, 1) * x_j


def test_inheritance():
    x = torch.randn(4, 8)
    edge_index = torch.tensor([[0, 1, 2, 3], [0, 0, 1, 1]])
    edge_weight = torch.rand(4)

    conv = EdgeGraphConv(8, 16)
    assert conv(x, edge_index, edge_weight).size() == (4, 16)



import pytest
import torch

import torch_geometric.typing
from torch_geometric.nn import MultiAggregation


@pytest.mark.parametrize('multi_aggr_tuple', [
    (dict(mode='cat'), 3),
    (dict(mode='proj', mode_kwargs=dict(in_channels=16, out_channels=16)), 1),
    (dict(mode='attn', mode_kwargs=dict(in_channels=16, out_channels=16,
                                        num_heads=4)), 1),
    (dict(mode='sum'), 1),
    (dict(mode='mean'), 1),
    (dict(mode='max'), 1),
    (dict(mode='min'), 1),
    (dict(mode='logsumexp'), 1),
    (dict(mode='std'), 1),
    (dict(mode='var'), 1),
])
def test_multi_aggr(multi_aggr_tuple):
    # The 'cat' combine mode will expand the output dimensions by
    # the number of aggregators which is 3 here, while the other
    # modes keep output dimensions unchanged.
    aggr_kwargs, expand = multi_aggr_tuple
    x = torch.randn(7, 16)
    index = torch.tensor([0, 0, 1, 1, 1, 2, 3])
    ptr = torch.tensor([0, 2, 5, 6, 7])

    aggrs = ['mean', 'sum', 'max']
    aggr = MultiAggregation(aggrs, **aggr_kwargs)
    aggr.reset_parameters()
    assert str(aggr) == ('MultiAggregation([\n'
                         '  MeanAggregation(),\n'
                         '  SumAggregation(),\n'
                         '  MaxAggregation(),\n'
                         f"], mode={aggr_kwargs['mode']})")

    out = aggr(x, index)
    assert out.size() == (4, expand * x.size(1))

    if (not torch_geometric.typing.WITH_TORCH_SCATTER
            and not torch_geometric.typing.WITH_PT20):
        with pytest.raises(ImportError, match="requires the 'torch-scatter'"):
            aggr(x, ptr=ptr)
    else:
        assert torch.allclose(out, aggr(x, ptr=ptr))

    jit = torch.jit.script(aggr)
    assert torch.allclose(out, jit(x, index))



import argparse
from itertools import product

from asap import ASAP
from datasets import get_dataset
from diff_pool import DiffPool
from edge_pool import EdgePool
from gcn import GCN, GCNWithJK
from gin import GIN, GIN0, GIN0WithJK, GINWithJK
from global_attention import GlobalAttentionNet
from graclus import Graclus
from graph_sage import GraphSAGE, GraphSAGEWithJK
from sag_pool import SAGPool
from set2set import Set2SetNet
from sort_pool import SortPool
from top_k import TopK
from train_eval import cross_validation_with_val_set

parser = argparse.ArgumentParser()
parser.add_argument('--epochs', type=int, default=100)
parser.add_argument('--batch_size', type=int, default=128)
parser.add_argument('--lr', type=float, default=0.01)
parser.add_argument('--lr_decay_factor', type=float, default=0.5)
parser.add_argument('--lr_decay_step_size', type=int, default=50)
args = parser.parse_args()

layers = [1, 2, 3, 4, 5]
hiddens = [16, 32, 64, 128]
datasets = ['MUTAG', 'PROTEINS', 'IMDB-BINARY', 'REDDIT-BINARY']  # , 'COLLAB']
nets = [
    GCNWithJK,
    GraphSAGEWithJK,
    GIN0WithJK,
    GINWithJK,
    Graclus,
    TopK,
    SAGPool,
    DiffPool,
    EdgePool,
    GCN,
    GraphSAGE,
    GIN0,
    GIN,
    GlobalAttentionNet,
    Set2SetNet,
    SortPool,
    ASAP,
]


def logger(info):
    fold, epoch = info['fold'] + 1, info['epoch']
    val_loss, test_acc = info['val_loss'], info['test_acc']
    print(f'{fold:02d}/{epoch:03d}: Val Loss: {val_loss:.4f}, '
          f'Test Accuracy: {test_acc:.3f}')


results = []
for dataset_name, Net in product(datasets, nets):
    best_result = (float('inf'), 0, 0)  # (loss, acc, std)
    print(f'--\n{dataset_name} - {Net.__name__}')
    for num_layers, hidden in product(layers, hiddens):
        dataset = get_dataset(dataset_name, sparse=Net != DiffPool)
        model = Net(dataset, num_layers, hidden)
        loss, acc, std = cross_validation_with_val_set(
            dataset,
            model,
            folds=10,
            epochs=args.epochs,
            batch_size=args.batch_size,
            lr=args.lr,
            lr_decay_factor=args.lr_decay_factor,
            lr_decay_step_size=args.lr_decay_step_size,
            weight_decay=0,
            logger=None,
        )
        if loss < best_result[0]:
            best_result = (loss, acc, std)

    desc = f'{best_result[1]:.3f} ± {best_result[2]:.3f}'
    print(f'Best result - {desc}')
    results += [f'{dataset_name} - {model}: {desc}']
results = '\n'.join(results)
print(f'--\n{results}')



import torch

from torch_geometric import seed_everything
from torch_geometric.data import Data
from torch_geometric.datasets import InfectionDataset
from torch_geometric.datasets.graph_generator import ERGraph, GraphGenerator


class DummyGraph(GraphGenerator):
    def __call__(self) -> Data:
        edge_index = torch.tensor([
            [0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9],
            [1, 0, 2, 1, 3, 2, 4, 3, 5, 4, 6, 5, 7, 6, 8, 7, 9, 8],
        ])
        return Data(num_nodes=10, edge_index=edge_index)


def test_infection_dataset():
    seed_everything(12345)
    graph_generator = DummyGraph()
    dataset = InfectionDataset(graph_generator, num_infected_nodes=2,
                               max_path_length=2)
    assert str(dataset) == ('InfectionDataset(1, '
                            'graph_generator=DummyGraph(), '
                            'num_infected_nodes=2, '
                            'max_path_length=2)')
    assert len(dataset) == 1

    data = dataset[0]
    assert len(data) == 4
    assert data.x.size() == (10, 2)
    assert data.x[:, 0].sum() == 8 and data.x[:, 1].sum() == 2
    assert torch.equal(data.edge_index, graph_generator().edge_index)
    assert data.y.size() == (10, )

    # With `seed=12345`, node 0 and node 7 will be infected:
    assert data.x[0].tolist() == [0, 1]  # First infected node.
    assert data.x[7].tolist() == [0, 1]  # Second infected node.
    assert data.y.tolist() == [0, 1, 2, 3, 3, 2, 1, 0, 1, 2]
    assert data.edge_mask.tolist() == [
        1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0
    ]


def test_infection_dataset_reproducibility():
    graph_generator = ERGraph(num_nodes=500, edge_prob=0.004)

    seed_everything(12345)
    dataset1 = InfectionDataset(graph_generator, num_infected_nodes=50,
                                max_path_length=5)

    seed_everything(12345)
    dataset2 = InfectionDataset(graph_generator, num_infected_nodes=50,
                                max_path_length=5)

    assert torch.equal(dataset1[0].edge_mask, dataset2[0].edge_mask)



import torch

from torch_geometric.data import Data
from torch_geometric.transforms import RandomJitter


def test_random_jitter():
    assert str(RandomJitter(0.1)) == 'RandomJitter(0.1)'

    pos = torch.tensor([[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]])

    data = Data(pos=pos)
    data = RandomJitter(0)(data)
    assert len(data) == 1
    assert torch.allclose(data.pos, pos)

    data = Data(pos=pos)
    data = RandomJitter(0.1)(data)
    assert len(data) == 1
    assert data.pos.min() >= -0.1
    assert data.pos.max() <= 0.1

    data = Data(pos=pos)
    data = RandomJitter([0.1, 1])(data)
    assert len(data) == 1
    assert data.pos[:, 0].min() >= -0.1
    assert data.pos[:, 0].max() <= 0.1
    assert data.pos[:, 1].min() >= -1
    assert data.pos[:, 1].max() <= 1



import torch
from tqdm import tqdm

from torch_geometric.nn import GAT, to_hetero


class HeteroGAT(torch.nn.Module):
    def __init__(self, metadata, hidden_channels, num_layers, output_channels,
                 num_heads):
        super().__init__()
        self.model = to_hetero(
            GAT((-1, -1), hidden_channels, num_layers, output_channels,
                add_self_loops=False, heads=num_heads), metadata)

    def forward(self, x_dict, edge_index_dict):
        return self.model(x_dict, edge_index_dict)

    @torch.no_grad()
    def inference(self, loader, device, progress_bar=False, **kwargs):
        self.model.eval()
        if progress_bar:
            loader = tqdm(loader, desc="Inference")
        for batch in loader:
            batch = batch.to(device)
            if 'adj_t' in batch:
                self.model(batch.x_dict, batch.adj_t_dict)
            else:
                self.model(batch.x_dict, batch.edge_index_dict)

    @torch.no_grad()
    def test(self, x, loader, device, progress_bar=False):
        self.model.eval()
        total_examples = total_correct = 0
        if progress_bar:
            loader = tqdm(loader, desc="Evaluate")
        for batch in loader:
            batch = batch.to(device)
            if 'adj_t' in batch:
                out = self.model(batch.x_dict, batch.adj_t_dict)
            else:
                out = self.model(batch.x_dict, batch.edge_index_dict)
            batch_size = batch['paper'].batch_size
            out = out['paper'][:batch_size]
            pred = out.argmax(dim=-1)

            total_examples += batch_size
            total_correct += int((pred == batch['paper'].y[:batch_size]).sum())

        return total_correct / total_examples



import pytest
import torch

import torch_geometric.typing
from torch_geometric.nn import HEATConv
from torch_geometric.testing import is_full_test
from torch_geometric.typing import SparseTensor


@pytest.mark.parametrize('concat', [True, False])
def test_heat_conv(concat):
    x = torch.randn(4, 8)
    edge_index = torch.tensor([[0, 1, 2, 3], [0, 0, 1, 1]])
    edge_attr = torch.randn((4, 2))
    node_type = torch.tensor([0, 0, 1, 2])
    edge_type = torch.tensor([0, 2, 1, 2])

    conv = HEATConv(in_channels=8, out_channels=16, num_node_types=3,
                    num_edge_types=3, edge_type_emb_dim=5, edge_dim=2,
                    edge_attr_emb_dim=6, heads=2, concat=concat)
    assert str(conv) == 'HEATConv(8, 16, heads=2)'

    out = conv(x, edge_index, node_type, edge_type, edge_attr)
    assert out.size() == (4, 32 if concat else 16)

    if torch_geometric.typing.WITH_TORCH_SPARSE:
        adj = SparseTensor.from_edge_index(edge_index, edge_attr, (4, 4))
        assert torch.allclose(conv(x, adj.t(), node_type, edge_type), out,
                              atol=1e-5)

    if is_full_test():
        jit = torch.jit.script(conv)
        assert torch.allclose(
            jit(x, edge_index, node_type, edge_type, edge_attr), out,
            atol=1e-5)

        if torch_geometric.typing.WITH_TORCH_SPARSE:
            assert torch.allclose(jit(x, adj.t(), node_type, edge_type), out,
                                  atol=1e-5)



import os
import tempfile
import warnings
from typing import Any
from unittest.mock import patch

import pytest
import torch

from torch_geometric import is_in_onnx_export, safe_onnx_export

# Global mock to prevent ANY real ONNX calls in tests
# This ensures no deprecation warnings or real ONNX issues
pytestmark = pytest.mark.filterwarnings("ignore::DeprecationWarning")


class SimpleModel(torch.nn.Module):
    """Simple model for testing ONNX export."""
    def __init__(self) -> None:
        super().__init__()
        self.linear = torch.nn.Linear(4, 2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.linear(x)


def test_is_in_onnx_export() -> None:
    """Test is_in_onnx_export function."""
    assert not is_in_onnx_export()


def test_safe_onnx_export_ci_resilient() -> None:
    """Test safe_onnx_export handles CI environment issues gracefully."""
    model = SimpleModel()
    x = torch.randn(3, 4)

    # Use mocking to prevent real ONNX calls and deprecation warnings
    with patch('torch.onnx.export', return_value=None) as mock_export:
        with tempfile.NamedTemporaryFile(suffix='.onnx', delete=False) as f:
            try:
                # Test with skip_on_error=True - should never fail
                result = safe_onnx_export(model, (x, ), f.name,
                                          skip_on_error=True)
                # Should always succeed with mocking
                assert result is True

                # Verify the mock was called correctly
                mock_export.assert_called_once()
                call_args = mock_export.call_args[0]
                assert call_args[0] is model
                assert isinstance(call_args[1], tuple)
                assert call_args[2] == f.name

            finally:
                if os.path.exists(f.name):
                    try:
                        os.unlink(f.name)
                    except (PermissionError, OSError):
                        pass  # Ignore file lock issues


def test_safe_onnx_export_success() -> None:
    """Test successful ONNX export with pure mocking."""
    model = SimpleModel()
    x = torch.randn(3, 4)

    # Use comprehensive mocking to avoid any real ONNX calls
    with patch('torch.onnx.export', return_value=None) as mock_export:
        with tempfile.NamedTemporaryFile(suffix='.onnx', delete=False) as f:
            try:
                # Test with tuple args - should succeed with mock
                result = safe_onnx_export(model, (x, ), f.name)
                assert result is True

                # Verify torch.onnx.export was called with correct args
                mock_export.assert_called()
                call_args = mock_export.call_args[0]
                assert call_args[0] is model  # model
                assert isinstance(call_args[1], tuple)  # args as tuple
                assert call_args[2] == f.name  # file path

                # Reset mock for second test
                mock_export.reset_mock()

                # Test with single tensor (should be converted to tuple)
                result = safe_onnx_export(model, x, f.name)
                assert result is True

                # Verify single tensor was converted to tuple
                call_args = mock_export.call_args[0]
                assert isinstance(call_args[1], tuple)

            finally:
                if os.path.exists(f.name):
                    try:
                        try:

                            os.unlink(f.name)

                        except (PermissionError, OSError):

                            pass
                    except (PermissionError, OSError):
                        pass


def test_safe_onnx_export_with_skip_on_error() -> None:
    """Test safe_onnx_export with skip_on_error=True."""
    model = SimpleModel()
    x = torch.randn(3, 4)

    # Mock torch.onnx.export to raise SerdeError
    with patch('torch.onnx.export') as mock_export:
        mock_export.side_effect = Exception(
            "onnx_ir.serde.SerdeError: allowzero")

        with tempfile.NamedTemporaryFile(suffix='.onnx', delete=False) as f:
            try:
                # Should return False instead of raising
                result = safe_onnx_export(model, (x, ), f.name,
                                          skip_on_error=True)
                assert result is False
            finally:
                if os.path.exists(f.name):
                    try:

                        os.unlink(f.name)

                    except (PermissionError, OSError):

                        pass


def test_serde_error_patterns() -> None:
    """Test detection of various SerdeError patterns."""
    model = SimpleModel()
    x = torch.randn(3, 4)

    error_patterns = [
        "onnx_ir.serde.SerdeError: allowzero attribute",
        "ValueError: Value out of range: 1", "serialize_model_into failed",
        "serialize_attribute_into failed"
    ]

    for error_msg in error_patterns:
        # Use multiple patch targets to ensure comprehensive mocking
        with patch('torch.onnx.export') as mock_export, \
             patch('torch_geometric._onnx.torch.onnx.export') as mock_export2:

            mock_export.side_effect = Exception(error_msg)
            mock_export2.side_effect = Exception(error_msg)

            with tempfile.NamedTemporaryFile(suffix='.onnx',
                                             delete=False) as f:
                try:
                    result = safe_onnx_export(model, (x, ), f.name,
                                              skip_on_error=True)
                    assert result is False
                finally:
                    if os.path.exists(f.name):
                        try:
                            try:

                                os.unlink(f.name)

                            except (PermissionError, OSError):

                                pass
                        except (PermissionError, OSError):
                            pass  # Ignore file lock issues


def test_non_serde_error_reraise() -> None:
    """Test that non-SerdeError exceptions are re-raised."""
    model = SimpleModel()
    x = torch.randn(3, 4)

    # Use comprehensive mocking to prevent real ONNX calls
    with patch('torch.onnx.export') as mock_export:
        mock_export.side_effect = ValueError("Some other error")

        with tempfile.NamedTemporaryFile(suffix='.onnx', delete=False) as f:
            try:
                with pytest.raises(ValueError, match="Some other error"):
                    safe_onnx_export(model, (x, ), f.name)
            finally:
                if os.path.exists(f.name):
                    try:

                        os.unlink(f.name)

                    except (PermissionError, OSError):

                        pass


def test_dynamo_fallback() -> None:
    """Test dynamo=False fallback strategy."""
    model = SimpleModel()
    x = torch.randn(3, 4)

    call_count = 0

    def mock_export_side_effect(*_args: Any, **kwargs: Any) -> None:
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            # First call fails
            raise Exception("onnx_ir.serde.SerdeError: allowzero")
        elif call_count == 2 and not kwargs.get('dynamo', True):
            # Second call succeeds with dynamo=False
            return None
        else:
            raise Exception("Unexpected call")

    with patch('torch.onnx.export', side_effect=mock_export_side_effect):
        with tempfile.NamedTemporaryFile(suffix='.onnx', delete=False) as f:
            try:
                result = safe_onnx_export(model, (x, ), f.name, dynamo=True)
                assert result is True
                assert call_count == 2
            finally:
                if os.path.exists(f.name):
                    try:

                        os.unlink(f.name)

                    except (PermissionError, OSError):

                        pass


def test_opset_fallback() -> None:
    """Test opset version fallback strategy."""
    model = SimpleModel()
    x = torch.randn(3, 4)

    call_count = 0

    def mock_export_side_effect(*_args: Any, **kwargs: Any) -> None:
        nonlocal call_count
        call_count += 1
        # Fail until we get to opset_version=17
        if kwargs.get('opset_version') == 17:
            # This call succeeds
            return None
        else:
            # All other calls fail
            raise Exception("onnx_ir.serde.SerdeError: allowzero")

    with patch('torch.onnx.export', side_effect=mock_export_side_effect):
        with tempfile.NamedTemporaryFile(suffix='.onnx', delete=False) as f:
            try:
                result = safe_onnx_export(model, (x, ), f.name,
                                          opset_version=18)
                # Should succeed when opset_version=17 is tried
                assert result is True
            finally:
                if os.path.exists(f.name):
                    try:
                        try:

                            os.unlink(f.name)

                        except (PermissionError, OSError):

                            pass
                    except (PermissionError, OSError):
                        pass


def test_all_strategies_fail() -> None:
    """Test when all workaround strategies fail."""
    model = SimpleModel()
    x = torch.randn(3, 4)

    with patch('torch.onnx.export') as mock_export:
        mock_export.side_effect = Exception(
            "onnx_ir.serde.SerdeError: allowzero")

        with tempfile.NamedTemporaryFile(suffix='.onnx', delete=False) as f:
            try:
                # Should raise RuntimeError when skip_on_error=False
                with pytest.raises(RuntimeError,
                                   match="Failed to export model to ONNX"):
                    safe_onnx_export(model, (x, ), f.name, skip_on_error=False)

                # Should return False when skip_on_error=True
                result = safe_onnx_export(model, (x, ), f.name,
                                          skip_on_error=True)
                assert result is False
            finally:
                if os.path.exists(f.name):
                    try:

                        os.unlink(f.name)

                    except (PermissionError, OSError):

                        pass


def test_pytest_environment_detection() -> None:
    """Test pytest environment detection for better error messages."""
    model = SimpleModel()
    x = torch.randn(3, 4)

    with patch('torch.onnx.export') as mock_export:
        mock_export.side_effect = Exception(
            "onnx_ir.serde.SerdeError: allowzero")

        # Set pytest environment variable
        with patch.dict(os.environ, {'PYTEST_CURRENT_TEST': 'test_something'}):
            with tempfile.NamedTemporaryFile(suffix='.onnx',
                                             delete=False) as f:
                try:
                    with pytest.raises(RuntimeError) as exc_info:
                        safe_onnx_export(model, (x, ), f.name,
                                         skip_on_error=False)

                    # Should contain pytest-specific guidance
                    assert "pytest environments" in str(exc_info.value)
                    assert "torch.jit.script()" in str(exc_info.value)
                finally:
                    if os.path.exists(f.name):
                        try:

                            os.unlink(f.name)

                        except (PermissionError, OSError):

                            pass


def test_warnings_emitted() -> None:
    """Test that appropriate warnings are emitted during workarounds."""
    model = SimpleModel()
    x = torch.randn(3, 4)

    call_count = 0

    def mock_export_side_effect(*_args: Any, **_kwargs: Any) -> None:
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            raise Exception("onnx_ir.serde.SerdeError: allowzero")
        elif call_count == 2:
            return None  # Success on dynamo fallback
        else:
            raise Exception("Unexpected call")

    with patch('torch.onnx.export', side_effect=mock_export_side_effect):
        with tempfile.NamedTemporaryFile(suffix='.onnx', delete=False) as f:
            try:
                with warnings.catch_warnings(record=True) as w:
                    warnings.simplefilter("always")
                    result = safe_onnx_export(model, (x, ), f.name,
                                              dynamo=True)

                    assert result is True
                    assert len(w) >= 2  # Initial error + dynamo fallback
                    assert any("allowzero boolean attribute bug" in str(
                        warning.message) for warning in w)
                    assert any(
                        "dynamo=False as workaround" in str(warning.message)
                        for warning in w)
            finally:
                if os.path.exists(f.name):
                    try:

                        os.unlink(f.name)

                    except (PermissionError, OSError):

                        pass


@pytest.mark.parametrize(
    "args_input",
    [
        torch.randn(3, 4),  # Single tensor
        (torch.randn(3, 4), ),  # Tuple with one tensor
        (torch.randn(3, 4), torch.randn(3, 2)),  # Tuple with multiple tensors
    ])
def test_args_conversion(args_input: Any) -> None:
    """Test that args are properly converted to tuple format."""
    model = SimpleModel()

    with patch('torch.onnx.export') as mock_export:
        mock_export.return_value = None

        with tempfile.NamedTemporaryFile(suffix='.onnx', delete=False) as f:
            try:
                result = safe_onnx_export(model, args_input, f.name)
                assert result is True

                # Check that torch.onnx.export was called with tuple args
                mock_export.assert_called_once()
                call_args = mock_export.call_args[0]
                assert isinstance(call_args[1], tuple)  # args should be tuple
            finally:
                if os.path.exists(f.name):
                    try:

                        os.unlink(f.name)

                    except (PermissionError, OSError):

                        pass



import pytest
import torch

import torch_geometric.typing
from torch_geometric.nn import FastRGCNConv, RGCNConv
from torch_geometric.testing import is_full_test, withCUDA, withDevice
from torch_geometric.typing import SparseTensor

classes = [RGCNConv, FastRGCNConv]
confs = [(None, None), (2, None), (None, 2)]


@withDevice
@pytest.mark.parametrize('conf', confs)
def test_rgcn_conv_equality(conf, device):
    num_bases, num_blocks = conf

    x1 = torch.randn(4, 4, device=device)
    edge_index = torch.tensor([
        [0, 1, 1, 2, 2, 3],
        [0, 0, 1, 0, 1, 1],
    ], device=device)
    edge_type = torch.tensor([0, 1, 1, 0, 0, 1], device=device)

    edge_index = torch.tensor([
        [0, 1, 1, 2, 2, 3, 0, 1, 1, 2, 2, 3],
        [0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 1, 1],
    ], device=device)
    edge_type = torch.tensor([0, 1, 1, 0, 0, 1, 2, 3, 3, 2, 2, 3],
                             device=device)

    torch.manual_seed(12345)
    conv1 = RGCNConv(4, 32, 4, num_bases, num_blocks, aggr='sum').to(device)

    torch.manual_seed(12345)
    conv2 = FastRGCNConv(4, 32, 4, num_bases, num_blocks,
                         aggr='sum').to(device)

    out1 = conv1(x1, edge_index, edge_type)
    out2 = conv2(x1, edge_index, edge_type)
    assert torch.allclose(out1, out2, atol=1e-2)

    if num_blocks is None:
        out1 = conv1(None, edge_index, edge_type)
        out2 = conv2(None, edge_index, edge_type)
        assert torch.allclose(out1, out2, atol=1e-2)


@withCUDA
@pytest.mark.parametrize('cls', classes)
@pytest.mark.parametrize('conf', confs)
def test_rgcn_conv_basic(cls, conf, device):
    num_bases, num_blocks = conf

    x1 = torch.randn(4, 4, device=device)
    x2 = torch.randn(2, 16, device=device)
    idx1 = torch.arange(4, device=device)
    idx2 = torch.arange(2, device=device)
    edge_index = torch.tensor([
        [0, 1, 1, 2, 2, 3],
        [0, 0, 1, 0, 1, 1],
    ], device=device)
    edge_type = torch.tensor([0, 1, 1, 0, 0, 1], device=device)

    conv = cls(4, 32, 2, num_bases, num_blocks, aggr='sum').to(device)
    assert str(conv) == f'{cls.__name__}(4, 32, num_relations=2)'

    out1 = conv(x1, edge_index, edge_type)
    assert out1.size() == (4, 32)

    if torch_geometric.typing.WITH_TORCH_SPARSE:
        adj = SparseTensor.from_edge_index(edge_index, edge_type, (4, 4))
        assert torch.allclose(conv(x1, adj.t()), out1, atol=1e-3)

    if num_blocks is None:
        out2 = conv(None, edge_index, edge_type)
        assert torch.allclose(conv(idx1, edge_index, edge_type), out2, 1e-3)
        assert out2.size() == (4, 32)
        if torch_geometric.typing.WITH_TORCH_SPARSE:
            assert torch.allclose(conv(None, adj.t()), out2, atol=1e-3)
            assert torch.allclose(conv(idx1, adj.t()), out2, atol=1e-3)

    if is_full_test():
        jit = torch.jit.script(conv)
        assert torch.allclose(jit(x1, edge_index, edge_type), out1, atol=1e-3)
        if num_blocks is None:
            assert torch.allclose(jit(idx1, edge_index, edge_type), out2,
                                  atol=1e-3)
            assert torch.allclose(jit(None, edge_index, edge_type), out2,
                                  atol=1e-3)

        if torch_geometric.typing.WITH_TORCH_SPARSE:
            assert torch.allclose(jit(x1, adj.t()), out1)
            if num_blocks is None:
                assert torch.allclose(jit(idx1, adj.t()), out2, atol=1e-3)
                assert torch.allclose(jit(None, adj.t()), out2, atol=1e-3)

    # Test bipartite message passing:
    conv = cls((4, 16), 32, 2, num_bases, num_blocks, aggr='sum').to(device)
    assert str(conv) == f'{cls.__name__}((4, 16), 32, num_relations=2)'

    out1 = conv((x1, x2), edge_index, edge_type)
    assert out1.size() == (2, 32)

    if torch_geometric.typing.WITH_TORCH_SPARSE:
        adj = SparseTensor.from_edge_index(edge_index, edge_type, (4, 2))
        assert torch.allclose(conv((x1, x2), adj.t()), out1, atol=1e-3)

    if num_blocks is None:
        out2 = conv((None, idx2), edge_index, edge_type)
        assert out2.size() == (2, 32)
        assert torch.allclose(conv((idx1, idx2), edge_index, edge_type), out2,
                              atol=1e-3)
        if torch_geometric.typing.WITH_TORCH_SPARSE:
            assert torch.allclose(conv((None, idx2), adj.t()), out2, atol=1e-3)
            assert torch.allclose(conv((idx1, idx2), adj.t()), out2, atol=1e-3)

    if is_full_test():
        jit = torch.jit.script(conv)
        assert torch.allclose(jit((x1, x2), edge_index, edge_type), out1,
                              atol=1e-3)
        if num_blocks is None:
            assert torch.allclose(jit((None, idx2), edge_index, edge_type),
                                  out2, atol=1e-3)
            assert torch.allclose(jit((idx1, idx2), edge_index, edge_type),
                                  out2, atol=1e-3)

        if torch_geometric.typing.WITH_TORCH_SPARSE:
            assert torch.allclose(jit((x1, x2), adj.t()), out1, atol=1e-3)
            if num_blocks is None:
                assert torch.allclose(jit((None, idx2), adj.t()), out2,
                                      atol=1e-3)
                assert torch.allclose(jit((idx1, idx2), adj.t()), out2,
                                      atol=1e-3)



import pytest
import torch
from torch.optim.lr_scheduler import ConstantLR, LambdaLR, ReduceLROnPlateau

import torch_geometric
from torch_geometric.nn.resolver import (
    activation_resolver,
    aggregation_resolver,
    lr_scheduler_resolver,
    normalization_resolver,
    optimizer_resolver,
)


def test_activation_resolver():
    assert isinstance(activation_resolver(torch.nn.ELU()), torch.nn.ELU)
    assert isinstance(activation_resolver(torch.nn.ReLU()), torch.nn.ReLU)
    assert isinstance(activation_resolver(torch.nn.PReLU()), torch.nn.PReLU)

    assert isinstance(activation_resolver('elu'), torch.nn.ELU)
    assert isinstance(activation_resolver('relu'), torch.nn.ReLU)
    assert isinstance(activation_resolver('prelu'), torch.nn.PReLU)


@pytest.mark.parametrize('aggr_tuple', [
    (torch_geometric.nn.MeanAggregation, 'mean'),
    (torch_geometric.nn.SumAggregation, 'sum'),
    (torch_geometric.nn.SumAggregation, 'add'),
    (torch_geometric.nn.MaxAggregation, 'max'),
    (torch_geometric.nn.MinAggregation, 'min'),
    (torch_geometric.nn.MulAggregation, 'mul'),
    (torch_geometric.nn.VarAggregation, 'var'),
    (torch_geometric.nn.StdAggregation, 'std'),
    (torch_geometric.nn.SoftmaxAggregation, 'softmax'),
    (torch_geometric.nn.PowerMeanAggregation, 'powermean'),
])
def test_aggregation_resolver(aggr_tuple):
    aggr_module, aggr_repr = aggr_tuple
    assert isinstance(aggregation_resolver(aggr_module()), aggr_module)
    assert isinstance(aggregation_resolver(aggr_repr), aggr_module)


def test_multi_aggregation_resolver():
    aggr = aggregation_resolver(None)
    assert aggr is None

    aggr = aggregation_resolver(['sum', 'mean', None])
    assert isinstance(aggr, torch_geometric.nn.MultiAggregation)
    assert len(aggr.aggrs) == 3
    assert isinstance(aggr.aggrs[0], torch_geometric.nn.SumAggregation)
    assert isinstance(aggr.aggrs[1], torch_geometric.nn.MeanAggregation)
    assert aggr.aggrs[2] is None


@pytest.mark.parametrize('norm_tuple', [
    (torch_geometric.nn.BatchNorm, 'batch', (16, )),
    (torch_geometric.nn.BatchNorm, 'batch_norm', (16, )),
    (torch_geometric.nn.InstanceNorm, 'instance_norm', (16, )),
    (torch_geometric.nn.LayerNorm, 'layer_norm', (16, )),
    (torch_geometric.nn.GraphNorm, 'graph_norm', (16, )),
    (torch_geometric.nn.GraphSizeNorm, 'graphsize_norm', ()),
    (torch_geometric.nn.PairNorm, 'pair_norm', ()),
    (torch_geometric.nn.MessageNorm, 'message_norm', ()),
    (torch_geometric.nn.DiffGroupNorm, 'diffgroup_norm', (16, 4)),
])
def test_normalization_resolver(norm_tuple):
    norm_module, norm_repr, norm_args = norm_tuple
    assert isinstance(normalization_resolver(norm_module(*norm_args)),
                      norm_module)
    assert isinstance(normalization_resolver(norm_repr, *norm_args),
                      norm_module)


def test_optimizer_resolver():
    params = [torch.nn.Parameter(torch.randn(1))]

    assert isinstance(optimizer_resolver(torch.optim.SGD(params, lr=0.01)),
                      torch.optim.SGD)
    assert isinstance(optimizer_resolver(torch.optim.Adam(params)),
                      torch.optim.Adam)
    assert isinstance(optimizer_resolver(torch.optim.Rprop(params)),
                      torch.optim.Rprop)

    assert isinstance(optimizer_resolver('sgd', params, lr=0.01),
                      torch.optim.SGD)
    assert isinstance(optimizer_resolver('adam', params), torch.optim.Adam)
    assert isinstance(optimizer_resolver('rprop', params), torch.optim.Rprop)


@pytest.mark.parametrize('scheduler_args', [
    ('constant_with_warmup', LambdaLR),
    ('linear_with_warmup', LambdaLR),
    ('cosine_with_warmup', LambdaLR),
    ('cosine_with_warmup_restarts', LambdaLR),
    ('polynomial_with_warmup', LambdaLR),
    ('constant', ConstantLR),
    ('ReduceLROnPlateau', ReduceLROnPlateau),
])
def test_lr_scheduler_resolver(scheduler_args):
    scheduler_name, scheduler_cls = scheduler_args

    model = torch.nn.Linear(10, 5)
    optimizer = torch.optim.Adam(model.parameters(), lr=0.01)

    lr_scheduler = lr_scheduler_resolver(
        scheduler_name,
        optimizer,
        num_training_steps=100,
    )
    assert isinstance(lr_scheduler, scheduler_cls)



import torch

from torch_geometric.nn.models import SGFormer


def test_sgformer():
    x = torch.randn(10, 16)
    edge_index = torch.tensor([
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        [1, 2, 3, 4, 0, 6, 7, 8, 9, 5],
    ])
    batch = torch.tensor([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])

    model = SGFormer(
        in_channels=16,
        hidden_channels=128,
        out_channels=40,
    )
    out = model(x, edge_index, batch)
    assert out.size() == (10, 40)



import pytest
import torch

from torch_geometric.explain import DummyExplainer, Explainer, Explanation
from torch_geometric.explain.config import ExplanationType


class DummyModel(torch.nn.Module):
    def forward(self, x, edge_index):
        return x.mean().view(-1)


def test_get_prediction(data):
    model = DummyModel()
    assert model.training

    explainer = Explainer(
        model,
        algorithm=DummyExplainer(),
        explanation_type='phenomenon',
        node_mask_type='object',
        model_config=dict(
            mode='regression',
            task_level='graph',
        ),
    )
    pred = explainer.get_prediction(data.x, data.edge_index)
    assert model.training
    assert pred.size() == (1, )


@pytest.mark.parametrize('target', [None, torch.randn(2)])
@pytest.mark.parametrize('explanation_type', [x for x in ExplanationType])
def test_forward(data, target, explanation_type):
    model = DummyModel()
    assert model.training

    explainer = Explainer(
        model,
        algorithm=DummyExplainer(),
        explanation_type=explanation_type,
        node_mask_type='attributes',
        model_config=dict(
            mode='regression',
            task_level='graph',
        ),
    )

    if target is None and explanation_type == ExplanationType.phenomenon:
        with pytest.raises(ValueError):
            explainer(data.x, data.edge_index, target=target)
    else:
        explanation = explainer(
            data.x,
            data.edge_index,
            target=target
            if explanation_type == ExplanationType.phenomenon else None,
        )
        assert model.training
        assert isinstance(explanation, Explanation)
        assert 'x' in explanation
        assert 'edge_index' in explanation
        assert 'target' in explanation
        assert 'node_mask' in explanation.available_explanations
        assert explanation.node_mask.size() == data.x.size()


@pytest.mark.parametrize('threshold_value', [0.2, 0.5, 0.8])
@pytest.mark.parametrize('node_mask_type', ['object', 'attributes'])
def test_hard_threshold(data, threshold_value, node_mask_type):
    explainer = Explainer(
        DummyModel(),
        algorithm=DummyExplainer(),
        explanation_type='model',
        node_mask_type=node_mask_type,
        edge_mask_type='object',
        model_config=dict(
            mode='regression',
            task_level='graph',
        ),
        threshold_config=('hard', threshold_value),
    )
    explanation = explainer(data.x, data.edge_index)

    assert 'node_mask' in explanation.available_explanations
    assert 'edge_mask' in explanation.available_explanations

    for key in explanation.available_explanations:
        mask = explanation[key]
        assert set(mask.unique().tolist()).issubset({0, 1})


@pytest.mark.parametrize('threshold_value', [1, 5, 10])
@pytest.mark.parametrize('threshold_type', ['topk', 'topk_hard'])
@pytest.mark.parametrize('node_mask_type', ['object', 'attributes'])
def test_topk_threshold(data, threshold_value, threshold_type, node_mask_type):
    explainer = Explainer(
        DummyModel(),
        algorithm=DummyExplainer(),
        explanation_type='model',
        node_mask_type=node_mask_type,
        edge_mask_type='object',
        model_config=dict(
            mode='regression',
            task_level='graph',
        ),
        threshold_config=(threshold_type, threshold_value),
    )
    explanation = explainer(data.x, data.edge_index)

    assert 'node_mask' in explanation.available_explanations
    assert 'edge_mask' in explanation.available_explanations

    for key in explanation.available_explanations:
        mask = explanation[key]
        if threshold_type == 'topk':
            assert (mask > 0).sum() == min(mask.numel(), threshold_value)
            assert ((mask == 0).sum() == mask.numel() -
                    min(mask.numel(), threshold_value))
        else:
            assert (mask == 1).sum() == min(mask.numel(), threshold_value)
            assert ((mask == 0).sum() == mask.numel() -
                    min(mask.numel(), threshold_value))



from typing import List

import pytest
import torch

from torch_geometric.metrics import (
    LinkPredAveragePopularity,
    LinkPredCoverage,
    LinkPredDiversity,
    LinkPredF1,
    LinkPredHitRatio,
    LinkPredMAP,
    LinkPredMetricCollection,
    LinkPredMRR,
    LinkPredNDCG,
    LinkPredPersonalization,
    LinkPredPrecision,
    LinkPredRecall,
)
from torch_geometric.testing import withCUDA


@pytest.mark.parametrize('num_src_nodes', [100])
@pytest.mark.parametrize('num_dst_nodes', [1000])
@pytest.mark.parametrize('num_edges', [3000])
@pytest.mark.parametrize('batch_size', [32])
@pytest.mark.parametrize('k', [1, 10, 100])
def test_precision(num_src_nodes, num_dst_nodes, num_edges, batch_size, k):
    row = torch.randint(0, num_src_nodes, (num_edges, ))
    col = torch.randint(0, num_dst_nodes, (num_edges, ))
    edge_label_index = torch.stack([row, col], dim=0)

    pred = torch.rand(num_src_nodes, num_dst_nodes)
    pred[row, col] += 0.3  # Offset positive links by a little.
    pred_index_mat = pred.topk(k, dim=1)[1]

    metric = LinkPredPrecision(k)
    assert str(metric) == f'LinkPredPrecision(k={k})'

    for node_id in torch.split(torch.randperm(num_src_nodes), batch_size):
        mask = torch.isin(edge_label_index[0], node_id)

        y_batch, y_index = edge_label_index[:, mask]
        # Remap `y_batch` back to `[0, batch_size - 1]` range:
        arange = torch.empty(num_src_nodes, dtype=node_id.dtype)
        arange[node_id] = torch.arange(node_id.numel())
        y_batch = arange[y_batch]

        metric.update(pred_index_mat[node_id], (y_batch, y_index))

    out = metric.compute()
    metric.reset()

    values: List[float] = []
    for i in range(num_src_nodes):  # Naive computation per node:
        y_index = col[row == i]
        if y_index.numel() > 0:
            mask = torch.isin(pred_index_mat[i], y_index)
            precision = float(mask.sum() / k)
            values.append(precision)
    expected = torch.tensor(values).mean()
    assert torch.allclose(out, expected)

    # Test with `k > pred_index_mat.size(1)`:
    metric.update(pred_index_mat[:, :k - 1], edge_label_index)
    metric.compute()
    metric.reset()


def test_recall():
    pred_index_mat = torch.tensor([[1, 0], [1, 2], [0, 2]])
    edge_label_index = torch.tensor([[0, 0, 0, 2, 2], [0, 1, 2, 2, 1]])
    edge_label_weight = torch.tensor([4.0, 1.0, 2.0, 3.0, 0.5])

    metric = LinkPredRecall(k=2)
    assert str(metric) == 'LinkPredRecall(k=2)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()
    assert float(result) == pytest.approx(0.5 * (2 / 3 + 0.5))

    # Test with `k > pred_index_mat.size(1)`:
    metric.update(pred_index_mat[:, :1], edge_label_index)
    metric.compute()
    metric.reset()

    metric = LinkPredRecall(k=2, weighted=True)
    assert str(metric) == 'LinkPredRecall(k=2, weighted=True)'
    with pytest.raises(ValueError, match="'edge_label_weight'"):
        metric.update(pred_index_mat, edge_label_index)

    metric.update(pred_index_mat, edge_label_index, edge_label_weight)
    result = metric.compute()
    metric.reset()
    assert float(result) == pytest.approx(0.5 * (5.0 / 7.0 + 3.0 / 3.5))

    edge_label_weight[0] = -2
    metric.update(pred_index_mat, edge_label_index, edge_label_weight)
    result = metric.compute()
    metric.reset()
    assert float(result) == pytest.approx(0.5 * (1.0 / 3.0 + 3.0 / 3.5))

    # Test with `k > pred_index_mat.size(1)`:
    metric.update(pred_index_mat[:, :1], edge_label_index, edge_label_weight)
    metric.compute()
    metric.reset()


def test_f1():
    pred_index_mat = torch.tensor([[1, 0], [1, 2], [0, 2]])
    edge_label_index = torch.tensor([[0, 0, 0, 2, 2], [0, 1, 2, 2, 1]])

    metric = LinkPredF1(k=2)
    assert str(metric) == 'LinkPredF1(k=2)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()
    assert float(result) == pytest.approx(0.6500)

    # Test with `k > pred_index_mat.size(1)`:
    metric.update(pred_index_mat[:, :1], edge_label_index)
    metric.compute()
    metric.reset()


def test_map():
    pred_index_mat = torch.tensor([[1, 0], [1, 2], [0, 2]])
    edge_label_index = torch.tensor([[0, 0, 0, 2, 2], [0, 1, 2, 2, 1]])

    metric = LinkPredMAP(k=2)
    assert str(metric) == 'LinkPredMAP(k=2)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()
    assert float(result) == pytest.approx(0.6250)

    # Test with `k > pred_index_mat.size(1)`:
    metric.update(pred_index_mat[:, :1], edge_label_index)
    metric.compute()
    metric.reset()


def test_ndcg():
    pred_index_mat = torch.tensor([[1, 0], [1, 2], [0, 2]])
    edge_label_index = torch.tensor([[0, 0, 0, 2, 2], [0, 1, 2, 2, 1]])
    edge_label_weight = torch.tensor([1.0, 2.0, 0.1, 3.0, 0.5])

    metric = LinkPredNDCG(k=2)
    assert str(metric) == 'LinkPredNDCG(k=2)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()
    assert float(result) == pytest.approx(0.6934264)

    # Test with `k > pred_index_mat.size(1)`:
    metric.update(pred_index_mat[:, :1], edge_label_index)
    metric.compute()
    metric.reset()

    metric = LinkPredNDCG(k=2, weighted=True)
    assert str(metric) == 'LinkPredNDCG(k=2, weighted=True)'
    with pytest.raises(ValueError, match="'edge_label_weight'"):
        metric.update(pred_index_mat, edge_label_index)

    metric.update(pred_index_mat, edge_label_index, edge_label_weight)
    result = metric.compute()
    metric.reset()
    assert float(result) == pytest.approx(0.7854486)

    perm = torch.randperm(edge_label_weight.size(0))
    metric.update(pred_index_mat, edge_label_index[:, perm],
                  edge_label_weight[perm])
    assert metric.compute() == result

    # Test with `k > pred_index_mat.size(1)`:
    metric.update(pred_index_mat[:, :1], edge_label_index, edge_label_weight)
    metric.compute()
    metric.reset()


def test_mrr():
    pred_index_mat = torch.tensor([[1, 0], [1, 2], [0, 2], [0, 1]])
    edge_label_index = torch.tensor([[0, 0, 2, 2, 3], [0, 1, 2, 1, 2]])

    metric = LinkPredMRR(k=2)
    assert str(metric) == 'LinkPredMRR(k=2)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()

    assert float(result) == pytest.approx((1 + 0.5 + 0) / 3)

    # Test with `k > pred_index_mat.size(1)`:
    metric.update(pred_index_mat[:, :1], edge_label_index)
    metric.compute()
    metric.reset()


def test_hit_ratio():
    pred_index_mat = torch.tensor([[1, 0], [1, 2], [0, 2], [0, 1]])
    edge_label_index = torch.tensor([[0, 0, 2, 2, 3], [0, 1, 2, 1, 2]])

    metric = LinkPredHitRatio(k=2)
    assert str(metric) == 'LinkPredHitRatio(k=2)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()

    assert float(result) == pytest.approx(2 / 3)

    # Test with `k > pred_index_mat.size(1)`:
    metric.update(pred_index_mat[:, :1], edge_label_index)
    metric.compute()
    metric.reset()


def test_coverage():
    pred_index_mat = torch.tensor([[1, 0], [1, 2], [0, 2], [0, 1]])
    edge_label_index = torch.empty(2, 0, dtype=torch.long)

    metric = LinkPredCoverage(k=2, num_dst_nodes=3)
    assert str(metric) == 'LinkPredCoverage(k=2, num_dst_nodes=3)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()
    metric.reset()
    assert metric.mask.sum() == 0

    assert float(result) == 1.0

    metric = LinkPredCoverage(k=1, num_dst_nodes=4)
    assert str(metric) == 'LinkPredCoverage(k=1, num_dst_nodes=4)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()
    metric.reset()
    assert metric.mask.sum() == 0

    assert float(result) == 2 / 4


def test_diversity():
    pred_index_mat = torch.tensor([[0, 1, 2], [3, 1, 0]])
    category = torch.tensor([0, 1, 2, 0])
    edge_label_index = torch.empty(2, 0, dtype=torch.long)

    metric = LinkPredDiversity(k=3, category=category)
    assert str(metric) == 'LinkPredDiversity(k=3)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()
    metric.reset()

    assert pytest.approx(float(result)) == (1 + 2 / 3) / 2


@withCUDA
def test_personalization(device):
    pred_index_mat = torch.tensor([[0, 1, 2, 3], [2, 1, 0, 4], [1, 0, 2, 5]],
                                  device=device)
    edge_label_index = torch.empty(2, 0, dtype=torch.long, device=device)

    metric = LinkPredPersonalization(k=4).to(device)
    assert str(metric) == 'LinkPredPersonalization(k=4)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()
    assert result.device == device
    assert float(result) == 0.25
    metric.reset()
    assert metric.preds == []

    metric.update(pred_index_mat[:0], edge_label_index)
    result = metric.compute()
    assert result.device == device
    assert float(result) == 0.0
    metric.reset()


def test_average_popularity():
    pred_index_mat = torch.tensor([[0, 1, 2], [3, 1, 0]])
    popularity = torch.tensor([10, 5, 2, 1])
    edge_label_index = torch.empty(2, 0, dtype=torch.long)

    metric = LinkPredAveragePopularity(k=3, popularity=popularity)
    assert str(metric) == 'LinkPredAveragePopularity(k=3)'
    metric.update(pred_index_mat, edge_label_index)
    result = metric.compute()
    metric.reset()

    assert pytest.approx(float(result)) == (10 + 5 + 2 + 1 + 5 + 10) / 6


@pytest.mark.parametrize('num_src_nodes', [10])
@pytest.mark.parametrize('num_dst_nodes', [50])
@pytest.mark.parametrize('num_edges', [200])
def test_metric_collection(num_src_nodes, num_dst_nodes, num_edges):
    metrics = [
        LinkPredMAP(k=10),
        LinkPredPrecision(k=100),
        LinkPredRecall(k=50),
        LinkPredF1(k=20),
        LinkPredMRR(k=40),
        LinkPredNDCG(k=80),
        LinkPredCoverage(k=5, num_dst_nodes=num_dst_nodes),
    ]

    row = torch.randint(0, num_src_nodes, (num_edges, ))
    col = torch.randint(0, num_dst_nodes, (num_edges, ))
    edge_label_index = torch.stack([row, col], dim=0)

    pred = torch.rand(num_src_nodes, num_dst_nodes)
    pred[row, col] += 0.3  # Offset positive links by a little.
    pred_index_mat = pred.argsort(dim=1)

    metric_collection = LinkPredMetricCollection(metrics)
    assert str(metric_collection) == (
        'LinkPredMetricCollection([\n'
        '  LinkPredMAP@10: LinkPredMAP(k=10),\n'
        '  LinkPredPrecision@100: LinkPredPrecision(k=100),\n'
        '  LinkPredRecall@50: LinkPredRecall(k=50),\n'
        '  LinkPredF1@20: LinkPredF1(k=20),\n'
        '  LinkPredMRR@40: LinkPredMRR(k=40),\n'
        '  LinkPredNDCG@80: LinkPredNDCG(k=80),\n'
        '  LinkPredCoverage@5: LinkPredCoverage(k=5, num_dst_nodes=50),\n'
        '])')
    assert metric_collection.max_k == 100

    expected = {}
    for metric in metrics:
        metric.update(pred_index_mat[:, :metric.k], edge_label_index)
        out = metric.compute()
        expected[f'{metric.__class__.__name__}@{metric.k}'] = out
        metric.reset()

    metric_collection.update(pred_index_mat, edge_label_index)
    assert metric_collection.compute() == expected
    metric_collection.reset()


def test_empty_ground_truth():
    pred = torch.rand(10, 5)
    pred_index_mat = pred.argsort(dim=1)
    edge_label_index = torch.empty(2, 0, dtype=torch.long)
    edge_label_weight = torch.empty(0)

    metric = LinkPredMAP(k=5)
    metric.update(pred_index_mat, edge_label_index)
    assert metric.compute() == 0
    metric.reset()

    metric = LinkPredNDCG(k=5, weighted=True)
    metric.update(pred_index_mat, edge_label_index, edge_label_weight)
    assert metric.compute() == 0
    metric.reset()



import torch
import torch.nn.functional as F

from torch_geometric.nn import GCNConv


class GCN(torch.nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.conv1 = GCNConv(in_channels, 16, cached=True)
        self.conv2 = GCNConv(16, out_channels, cached=True)

    def forward(self, data):
        x, edge_index = data.x, data.edge_index
        x = F.relu(self.conv1(x, edge_index))
        x = F.dropout(x, training=self.training)
        x = self.conv2(x, edge_index)
        return F.log_softmax(x, dim=1)



"""Distributed GAT training, targeting XPU devices.
PVC has 2 tiles, each reports itself as a separate
device. DDP approach allows us to employ both tiles.

Additional requirements:
    IPEX (intel_extension_for_pytorch)
    oneCCL (oneccl_bindings_for_pytorch)

    We need to import both these modules, as they extend
    torch module with XPU/oneCCL related functionality.

Run with:
    mpirun -np 2 python distributed_sampling_xpu.py
"""

import copy
import os
import os.path as osp
from typing import Any, Tuple, Union

import intel_extension_for_pytorch  # noqa
import oneccl_bindings_for_pytorch  # noqa
import torch
import torch.distributed as dist
import torch.nn.functional as F
from ogb.nodeproppred import Evaluator, PygNodePropPredDataset
from torch import Tensor
from torch.nn import Linear as Lin
from torch.nn.parallel import DistributedDataParallel as DDP
from tqdm import tqdm

from torch_geometric.loader import NeighborLoader
from torch_geometric.nn import GATConv
from torch_geometric.profile import get_stats_summary, profileit


class GAT(torch.nn.Module):
    def __init__(
        self,
        in_channels: int,
        hidden_channels: int,
        out_channels: int,
        num_layers: int,
        heads: int,
    ):
        super().__init__()

        self.num_layers = num_layers

        self.convs = torch.nn.ModuleList()
        self.convs.append(GATConv(dataset.num_features, hidden_channels,
                                  heads))
        for _ in range(num_layers - 2):
            self.convs.append(
                GATConv(heads * hidden_channels, hidden_channels, heads))
        self.convs.append(
            GATConv(heads * hidden_channels, out_channels, heads,
                    concat=False))

        self.skips = torch.nn.ModuleList()
        self.skips.append(Lin(dataset.num_features, hidden_channels * heads))
        for _ in range(num_layers - 2):
            self.skips.append(
                Lin(hidden_channels * heads, hidden_channels * heads))
        self.skips.append(Lin(hidden_channels * heads, out_channels))

    def forward(self, x: Tensor, edge_index: Tensor) -> Tensor:
        for i, (conv, skip) in enumerate(zip(self.convs, self.skips)):
            x = conv(x, edge_index) + skip(x)
            if i != self.num_layers - 1:
                x = F.elu(x)
                x = F.dropout(x, p=0.5, training=self.training)
        return x

    def inference(
        self,
        x_all: Tensor,
        device: Union[str, torch.device],
        subgraph_loader: NeighborLoader,
    ) -> Tensor:
        pbar = tqdm(total=x_all.size(0) * self.num_layers)
        pbar.set_description("Evaluating")

        # Compute representations of nodes layer by layer, using *all*
        # available edges. This leads to faster computation in contrast to
        # immediately computing the final representations of each batch.
        for i in range(self.num_layers):
            xs = []
            for batch in subgraph_loader:
                x = x_all[batch.n_id].to(device)
                edge_index = batch.edge_index.to(device)
                x = self.convs[i](x, edge_index) + self.skips[i](x)
                x = x[:batch.batch_size]
                if i != self.num_layers - 1:
                    x = F.elu(x)
                xs.append(x.cpu())

                pbar.update(batch.batch_size)

            x_all = torch.cat(xs, dim=0)

        pbar.close()

        return x_all


@profileit('xpu')
def train_step(model: Any, optimizer: Any, x: Tensor, edge_index: Tensor,
               y: Tensor, bs: int) -> float:
    optimizer.zero_grad()
    out = model(x, edge_index)[:bs]
    loss = F.cross_entropy(out, y[:bs].squeeze())
    loss.backward()
    optimizer.step()
    return float(loss)


def run(rank: int, world_size: int, dataset: PygNodePropPredDataset):
    device = f"xpu:{rank}"

    split_idx = dataset.get_idx_split()
    split_idx["train"] = (split_idx["train"].split(
        split_idx["train"].size(0) // world_size, dim=0)[rank].clone())
    data = dataset[0].to(device, "x", "y")

    kwargs = dict(batch_size=1024, num_workers=0, pin_memory=True)
    train_loader = NeighborLoader(data, input_nodes=split_idx["train"],
                                  num_neighbors=[10, 10, 5], **kwargs)

    if rank == 0:
        subgraph_loader = NeighborLoader(copy.copy(data), num_neighbors=[-1],
                                         **kwargs)
        evaluator = Evaluator(name="ogbn-products")

    torch.manual_seed(12345)
    model = GAT(dataset.num_features, 128, dataset.num_classes, num_layers=3,
                heads=4).to(device)
    model = DDP(model, device_ids=[device])
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)

    for epoch in range(1, 21):
        epoch_stats = []
        model.train()
        for batch in train_loader:
            batch = batch.to(device)
            loss, stats = train_step(model, optimizer, batch.x,
                                     batch.edge_index, batch.y,
                                     batch.batch_size)
            epoch_stats.append(stats)

        dist.barrier()

        if rank == 0:
            print(f"Epoch: {epoch:02d}, Loss: {loss:.4f}")

        print(f"Epoch: {epoch:02d}, Rank: {rank}, "
              f"Stats: {get_stats_summary(epoch_stats)}")

        if rank == 0 and epoch % 5 == 0:  # Evaluation on a single GPU
            model.eval()
            with torch.no_grad():
                out = model.module.inference(data.x, device, subgraph_loader)

            y_true = data.y.to(out.device)
            y_pred = out.argmax(dim=-1, keepdim=True)

            train_acc = evaluator.eval({
                "y_true": y_true[split_idx["train"]],
                "y_pred": y_pred[split_idx["train"]],
            })["acc"]
            val_acc = evaluator.eval({
                "y_true": y_true[split_idx["valid"]],
                "y_pred": y_pred[split_idx["valid"]],
            })["acc"]
            test_acc = evaluator.eval({
                "y_true": y_true[split_idx["test"]],
                "y_pred": y_pred[split_idx["test"]],
            })["acc"]

            print(f"Train: {train_acc:.4f}, Val: {val_acc:.4f}, "
                  f"Test: {test_acc:.4f}")

        dist.barrier()

    dist.destroy_process_group()


def get_dist_params() -> Tuple[int, int, str]:
    master_addr = "127.0.0.1"
    master_port = "29500"
    os.environ["MASTER_ADDR"] = master_addr
    os.environ["MASTER_PORT"] = master_port

    mpi_rank = int(os.environ.get("PMI_RANK", -1))
    mpi_world_size = int(os.environ.get("PMI_SIZE", -1))
    rank = mpi_rank if mpi_world_size > 0 else os.environ.get("RANK", 0)
    world_size = (mpi_world_size if mpi_world_size > 0 else os.environ.get(
        "WORLD_SIZE", 1))

    os.environ["RANK"] = str(rank)
    os.environ["WORLD_SIZE"] = str(world_size)

    init_method = f"tcp://{master_addr}:{master_port}"

    return rank, world_size, init_method


if __name__ == "__main__":
    rank, world_size, init_method = get_dist_params()
    dist.init_process_group(backend="ccl", init_method=init_method,
                            world_size=world_size, rank=rank)

    path = osp.join(osp.dirname(osp.realpath(__file__)), "../../data",
                    "ogbn-products")
    dataset = PygNodePropPredDataset("ogbn-products", path)

    run(rank, world_size, dataset)



import pytest
import torch

from torch_geometric import EdgeIndex
from torch_geometric.nn import CuGraphRGCNConv
from torch_geometric.nn import FastRGCNConv as RGCNConv
from torch_geometric.testing import onlyCUDA, withPackage


@onlyCUDA
@withPackage('pylibcugraphops>=23.02')
@pytest.mark.parametrize('aggr', ['add', 'sum', 'mean'])
@pytest.mark.parametrize('bias', [True, False])
@pytest.mark.parametrize('bipartite', [True, False])
@pytest.mark.parametrize('max_num_neighbors', [8, None])
@pytest.mark.parametrize('num_bases', [1, 2, None])
@pytest.mark.parametrize('root_weight', [True, False])
def test_rgcn_conv_equality(aggr, bias, bipartite, max_num_neighbors,
                            num_bases, root_weight):

    in_channels, out_channels, num_relations = (4, 2, 3)
    kwargs = dict(aggr=aggr, bias=bias, num_bases=num_bases,
                  root_weight=root_weight)

    size = (10, 8) if bipartite else (10, 10)
    x = torch.rand(size[0], in_channels, device='cuda')
    edge_index = torch.tensor([
        [7, 0, 0, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 8, 9],
        [0, 1, 2, 3, 4, 5, 6, 7, 7, 7, 7, 7, 7, 7, 7],
    ], device='cuda')
    edge_type = torch.tensor([1, 2, 1, 0, 2, 1, 2, 0, 2, 2, 1, 1, 1, 2,
                              2]).cuda()

    torch.manual_seed(12345)
    conv1 = RGCNConv(in_channels, out_channels, num_relations, **kwargs).cuda()
    torch.manual_seed(12345)
    conv2 = CuGraphRGCNConv(in_channels, out_channels, num_relations,
                            **kwargs).cuda()

    if bipartite:
        out1 = conv1((x, x[:size[1]]), edge_index, edge_type)
    else:
        out1 = conv1(x, edge_index, edge_type)

    out2 = conv2(
        x,
        EdgeIndex(edge_index, sparse_size=size),
        edge_type,
        max_num_neighbors=max_num_neighbors,
    )
    assert torch.allclose(out1, out2, atol=1e-3)

    grad_out = torch.rand_like(out1)
    out1.backward(grad_out)
    out2.backward(grad_out)

    end = -1 if root_weight else None
    assert torch.allclose(conv1.weight.grad, conv2.weight.grad[:end],
                          atol=1e-3)

    if root_weight:
        assert torch.allclose(conv1.root.grad, conv2.weight.grad[-1],
                              atol=1e-3)

    if num_bases is not None:
        assert torch.allclose(conv1.comp.grad, conv2.comp.grad, atol=1e-3)



import argparse

import torch
import torch.nn.functional as F
from points.datasets import get_dataset
from points.train_eval import run
from torch.nn import Linear as Lin

from torch_geometric.nn import XConv, fps, global_mean_pool
from torch_geometric.profile import rename_profile_file

parser = argparse.ArgumentParser()
parser.add_argument('--epochs', type=int, default=200)
parser.add_argument('--batch_size', type=int, default=32)
parser.add_argument('--lr', type=float, default=0.001)
parser.add_argument('--lr_decay_factor', type=float, default=0.5)
parser.add_argument('--lr_decay_step_size', type=int, default=50)
parser.add_argument('--weight_decay', type=float, default=0)
parser.add_argument('--inference', action='store_true')
parser.add_argument('--profile', action='store_true')
parser.add_argument('--bf16', action='store_true')
parser.add_argument('--compile', action='store_true')
args = parser.parse_args()


class Net(torch.nn.Module):
    def __init__(self, num_classes):
        super().__init__()

        self.conv1 = XConv(0, 48, dim=3, kernel_size=8, hidden_channels=32)
        self.conv2 = XConv(48, 96, dim=3, kernel_size=12, hidden_channels=64,
                           dilation=2)
        self.conv3 = XConv(96, 192, dim=3, kernel_size=16, hidden_channels=128,
                           dilation=2)
        self.conv4 = XConv(192, 384, dim=3, kernel_size=16,
                           hidden_channels=256, dilation=2)

        self.lin1 = Lin(384, 256)
        self.lin2 = Lin(256, 128)
        self.lin3 = Lin(128, num_classes)

    def forward(self, pos, batch):
        x = F.relu(self.conv1(None, pos, batch))

        idx = fps(pos, batch, ratio=0.375)
        x, pos, batch = x[idx], pos[idx], batch[idx]

        x = F.relu(self.conv2(x, pos, batch))

        idx = fps(pos, batch, ratio=0.334)
        x, pos, batch = x[idx], pos[idx], batch[idx]

        x = F.relu(self.conv3(x, pos, batch))
        x = F.relu(self.conv4(x, pos, batch))

        x = global_mean_pool(x, batch)

        x = F.relu(self.lin1(x))
        x = F.relu(self.lin2(x))
        x = F.dropout(x, p=0.5, training=self.training)
        x = self.lin3(x)
        return F.log_softmax(x, dim=-1)


train_dataset, test_dataset = get_dataset(num_points=1024)
model = Net(train_dataset.num_classes)
run(train_dataset, test_dataset, model, args.epochs, args.batch_size, args.lr,
    args.lr_decay_factor, args.lr_decay_step_size, args.weight_decay,
    args.inference, args.profile, args.bf16, args.compile)

if args.profile:
    rename_profile_file('points', XConv.__name__)



import torch

from torch_geometric.nn import PANConv, PANPooling
from torch_geometric.testing import is_full_test, withPackage


@withPackage('torch_sparse')
def test_pan_pooling():
    edge_index = torch.tensor([[0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3],
                               [1, 2, 3, 0, 2, 3, 0, 1, 3, 0, 1, 2]])
    num_nodes = edge_index.max().item() + 1
    x = torch.randn((num_nodes, 16))

    conv = PANConv(16, 32, filter_size=2)
    pool = PANPooling(32, ratio=0.5)
    assert str(pool) == 'PANPooling(32, ratio=0.5, multiplier=1.0)'

    x, M = conv(x, edge_index)
    h, edge_index, edge_weight, batch, perm, score = pool(x, M)

    assert h.size() == (2, 32)
    assert edge_index.size() == (2, 4)
    assert edge_weight.size() == (4, )
    assert perm.size() == (2, )
    assert score.size() == (2, )

    if is_full_test():
        jit = torch.jit.script(pool)
        out = jit(x, M)
        assert torch.allclose(h, out[0])
        assert torch.equal(edge_index, out[1])
        assert torch.allclose(edge_weight, out[2])
        assert torch.equal(batch, out[3])
        assert torch.equal(perm, out[4])
        assert torch.allclose(score, out[5])



import pytest
import torch

from torch_geometric.nn import GCNConv, Linear
from torch_geometric.utils import get_embeddings
from torch_geometric.utils.embedding import get_embeddings_hetero


class GNN(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = GCNConv(5, 6)
        self.conv2 = GCNConv(6, 7)

    def forward(self, x0, edge_index):
        x1 = self.conv1(x0, edge_index)
        x2 = self.conv2(x1, edge_index)
        return [x1, x2]


def test_get_embeddings():
    x = torch.randn(6, 5)
    edge_index = torch.tensor([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    with pytest.warns(UserWarning, match="any 'MessagePassing' layers"):
        intermediate_outs = get_embeddings(Linear(5, 5), x)
    assert len(intermediate_outs) == 0

    model = GNN()
    expected_embeddings = model(x, edge_index)

    embeddings = get_embeddings(model, x, edge_index)
    assert len(embeddings) == 2
    for expected, out in zip(expected_embeddings, embeddings):
        assert torch.allclose(expected, out)


def test_get_embeddings_hetero(hetero_data, hetero_model):
    # Create model using the metadata from hetero_data
    metadata = hetero_data.metadata()
    model = hetero_model(metadata)

    # Get heterogeneous embeddings
    embeddings_dict = get_embeddings_hetero(model, None, hetero_data.x_dict,
                                            hetero_data.edge_index_dict)

    # Verify the structure of the returned embeddings
    assert isinstance(embeddings_dict, dict)
    assert 'paper' in embeddings_dict
    assert 'author' in embeddings_dict

    # Verify that we have embeddings for both node types
    assert len(embeddings_dict['paper']) > 0
    assert len(embeddings_dict['author']) > 0

    # Check that the embeddings have the right shape
    num_paper_nodes = hetero_data['paper'].num_nodes
    num_author_nodes = hetero_data['author'].num_nodes

    # Verify dimensions of embeddings
    assert embeddings_dict['paper'][0].shape == (num_paper_nodes, 32
                                                 )  # First layer
    assert embeddings_dict['author'][0].shape == (num_author_nodes, 32
                                                  )  # First layer



import os.path as osp
import time

import torch
import torch.nn.functional as F

import torch_geometric.transforms as T
from torch_geometric.datasets import Planetoid
from torch_geometric.nn import GCNConv

if torch.cuda.is_available():
    device = torch.device('cuda')
elif hasattr(torch.backends, 'mps') and torch.backends.mps.is_available():
    device = torch.device('mps')
else:
    device = torch.device('cpu')

path = osp.dirname(osp.realpath(__file__))
path = osp.join(path, '..', '..', 'data', 'Planetoid')
dataset = Planetoid(
    path, name='Cora', transform=T.Compose([
        T.NormalizeFeatures(),
        T.GCNNorm(),
    ]))
data = dataset[0].to(device)


class GCN(torch.nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels):
        super().__init__()
        # Pre-process normalization to avoid CPU communication/graph breaks:
        self.conv1 = GCNConv(in_channels, hidden_channels, normalize=False)
        self.conv2 = GCNConv(hidden_channels, out_channels, normalize=False)

    def forward(self, x, edge_index, edge_weight):
        x = F.dropout(x, p=0.5, training=self.training)
        x = self.conv1(x, edge_index, edge_weight).relu()
        x = F.dropout(x, p=0.5, training=self.training)
        x = self.conv2(x, edge_index)
        return x


model = GCN(
    in_channels=dataset.num_features,
    hidden_channels=16,
    out_channels=dataset.num_classes,
).to(device)

# Compile the model into an optimized version:
model = torch.compile(model, dynamic=False)

optimizer = torch.optim.Adam([
    dict(params=model.conv1.parameters(), weight_decay=5e-4),
    dict(params=model.conv2.parameters(), weight_decay=0)
], lr=0.01)  # Only perform weight-decay on first convolution.


def train():
    model.train()
    optimizer.zero_grad()
    out = model(data.x, data.edge_index, data.edge_weight)
    loss = F.cross_entropy(out[data.train_mask], data.y[data.train_mask])
    loss.backward()
    optimizer.step()
    return float(loss.detach())


@torch.no_grad()
def test():
    model.eval()
    pred = model(data.x, data.edge_index, data.edge_weight).argmax(dim=-1)

    accs = []
    for mask in [data.train_mask, data.val_mask, data.test_mask]:
        accs.append(int((pred[mask] == data.y[mask]).sum()) / int(mask.sum()))
    return accs


times = []
for epoch in range(1, 201):
    start = time.time()
    loss = train()
    train_acc, val_acc, test_acc = test()
    times.append(time.time() - start)
    print(f'Epoch: {epoch:03d}, Loss: {loss:.4f}, Train: {train_acc:.4f}, '
          f'Val: {val_acc:.4f}, Test: {test_acc:.4f}')
print(f'Median time per epoch: {torch.tensor(times).median():.4f}s')



import argparse

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from sklearn.metrics import accuracy_score, average_precision_score, r2_score
from tqdm import tqdm

from torch_geometric.datasets import GraphLandDataset
from torch_geometric.nn import GCNConv


class Model(torch.nn.Module):
    def __init__(
        self,
        in_channels: int,
        hidden_channels: int,
        out_channels: int,
    ) -> None:
        super().__init__()
        self.conv = GCNConv(in_channels, hidden_channels)
        self.head = nn.Sequential(
            nn.ReLU(),
            nn.Linear(hidden_channels, out_channels),
        )

    def forward(
        self,
        x: torch.Tensor,
        edge_index: torch.Tensor,
    ) -> torch.Tensor:
        return self.head(self.conv(x, edge_index))


def _get_num_classes(dataset: GraphLandDataset) -> int:
    assert dataset.task != 'regression'
    targets = torch.cat([data.y for data in dataset], dim=0)
    return len(torch.unique(targets[~torch.isnan(targets)]))


def _train_step(
    model: nn.Module,
    dataset: GraphLandDataset,
    optimizer: optim.Optimizer,
) -> torch.Tensor:
    data = dataset[0]
    mask = data.train_mask if dataset.split != 'THI' else data.mask
    optimizer.zero_grad()
    outputs = model(data.x, data.edge_index).squeeze()

    if dataset.task == 'regression':
        loss = F.mse_loss(outputs[mask], data.y[mask])
    else:
        loss = F.cross_entropy(outputs[mask], data.y[mask].long())

    loss.backward()
    optimizer.step()
    return loss


def _eval_step(
    model: nn.Module,
    dataset: GraphLandDataset,
) -> dict[str, float]:
    def _compute_metric(outputs: np.ndarray, targets: np.ndarray) -> float:
        if dataset.task == 'regression':
            return float(r2_score(targets, outputs))

        elif dataset.task == 'binary_classification':
            predictions = outputs[:, 1]
            return float(average_precision_score(targets, predictions))

        else:
            predictions = np.argmax(outputs, axis=1)
            return float(accuracy_score(targets, predictions))

    metrics = dict()
    for idx, part in enumerate(['train', 'val', 'test']):
        if dataset.split == 'THI':
            data = dataset[idx]
            mask = data.mask
        else:
            data = dataset[0]
            mask = getattr(data, f'{part}_mask')

        outputs = model(data.x, data.edge_index).squeeze()
        metrics[part] = _compute_metric(
            outputs[mask].detach().cpu().numpy(),
            data.y[mask].cpu().numpy(),
        )
    return metrics


def _format_metrics(metrics: dict[str, float]) -> str:
    return ', '.join(f'{part}={metrics[part] * 100.0:.2f}'
                     for part in ['train', 'val', 'test'])


def run_experiment(name: str, split: str) -> None:
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    n_steps = 100
    dataset = GraphLandDataset(
        root='./data',
        split=split,
        name=name,
        to_undirected=True,
    ).to(device)
    model = Model(
        in_channels=dataset[0].x.shape[1],
        hidden_channels=256,
        out_channels=(_get_num_classes(dataset)
                      if dataset.task != 'regression' else 1),
    ).to(device)
    optimizer = optim.Adam(model.parameters(), lr=1e-3)

    best_metrics = {part: -float('inf') for part in ['train', 'val', 'test']}
    pbar = tqdm(range(n_steps))
    for _ in pbar:
        loss = _train_step(model, dataset, optimizer)
        curr_metrics = _eval_step(model, dataset)
        pbar.set_postfix_str(f'loss={loss.detach().cpu().item():.4f}, ' +
                             _format_metrics(curr_metrics))
        if curr_metrics['val'] > best_metrics['val']:
            best_metrics = curr_metrics

    print('Best metrics: ' + _format_metrics(best_metrics))
    return best_metrics


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--name',
        choices=list(GraphLandDataset.GRAPHLAND_DATASETS.keys()),
        help='The name of dataset.',
        required=True,
    )
    parser.add_argument(
        '--split',
        choices=['RL', 'RH', 'TH', 'THI'],
        help='The type of data split.',
        required=True,
    )
    args = parser.parse_args()
    run_experiment(args.name, args.split)



import torch
from torch.nn import Linear as Lin
from torch.nn import ReLU
from torch.nn import Sequential as Seq

from torch_geometric.nn.inits import (
    glorot,
    glorot_orthogonal,
    ones,
    reset,
    uniform,
    zeros,
)


def test_inits():
    x = torch.empty(1, 4)

    uniform(size=4, value=x)
    assert x.min() >= -0.5
    assert x.max() <= 0.5

    glorot(x)
    assert x.min() >= -1.1
    assert x.max() <= 1.1

    glorot_orthogonal(x, scale=1.0)
    assert x.min() >= -2.5
    assert x.max() <= 2.5

    zeros(x)
    assert x.tolist() == [[0, 0, 0, 0]]

    ones(x)
    assert x.tolist() == [[1, 1, 1, 1]]

    nn = Lin(16, 16)
    uniform(size=4, value=nn.weight)
    assert nn.weight[0].min() >= -0.5
    assert nn.weight[0].max() <= 0.5

    glorot(nn.weight)
    assert nn.weight[0].min() >= -0.45
    assert nn.weight[0].max() <= 0.45

    glorot_orthogonal(nn.weight, scale=1.0)
    assert nn.weight[0].min() >= -2.5
    assert nn.weight[0].max() <= 2.5


def test_reset():
    nn = Lin(16, 16)
    w = nn.weight.clone()
    reset(nn)
    assert not torch.allclose(nn.weight, w)

    nn = Seq(Lin(16, 16), ReLU(), Lin(16, 16))
    w_1, w_2 = nn[0].weight.clone(), nn[2].weight.clone()
    reset(nn)
    assert not torch.allclose(nn[0].weight, w_1)
    assert not torch.allclose(nn[2].weight, w_2)



from itertools import product

import torch

from torch_geometric.nn import dense_diff_pool
from torch_geometric.profile import benchmark
from torch_geometric.testing import is_full_test


def test_dense_diff_pool():
    batch_size, num_nodes, channels, num_clusters = (2, 20, 16, 10)
    x = torch.randn((batch_size, num_nodes, channels))
    adj = torch.rand((batch_size, num_nodes, num_nodes))
    s = torch.randn((batch_size, num_nodes, num_clusters))
    mask = torch.randint(0, 2, (batch_size, num_nodes), dtype=torch.bool)

    x_out, adj_out, link_loss, ent_loss = dense_diff_pool(x, adj, s, mask)
    assert x_out.size() == (2, 10, 16)
    assert adj_out.size() == (2, 10, 10)
    assert link_loss.item() >= 0
    assert ent_loss.item() >= 0

    if is_full_test():
        jit = torch.jit.script(dense_diff_pool)
        x_jit, adj_jit, link_loss, ent_loss = jit(x, adj, s, mask)
        assert torch.allclose(x_jit, x_out)
        assert torch.allclose(adj_jit, adj_out)
        assert link_loss.item() >= 0
        assert ent_loss.item() >= 0


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--device', type=str, default='cuda')
    parser.add_argument('--backward', action='store_true')
    args = parser.parse_args()

    BS = [2**i for i in range(4, 8)]
    NS = [2**i for i in range(4, 8)]
    FS = [2**i for i in range(5, 9)]
    CS = [2**i for i in range(5, 9)]

    funcs = []
    func_names = []
    args_list = []
    for B, N, F, C in product(BS, NS, FS, CS):
        x = torch.randn(B, N, F, device=args.device)
        adj = torch.randint(0, 2, (B, N, N), dtype=x.dtype, device=args.device)
        s = torch.randn(B, N, C, device=args.device)

        funcs.append(dense_diff_pool)
        func_names.append(f'B={B}, N={N}, F={F}, C={C}')
        args_list.append((x, adj, s))

    benchmark(
        funcs=funcs,
        func_names=func_names,
        args=args_list,
        num_steps=50 if args.device == 'cpu' else 500,
        num_warmups=10 if args.device == 'cpu' else 100,
        backward=args.backward,
        progress_bar=True,
    )



import os.path as osp

import pytest
import torch

from torch_geometric.testing import onlyGraphviz, withPackage
from torch_geometric.visualization import visualize_graph


@onlyGraphviz
@pytest.mark.parametrize('backend', [None, 'graphviz'])
def test_visualize_graph_via_graphviz(tmp_path, backend):
    edge_index = torch.tensor([
        [0, 1, 1, 2, 2, 3, 3, 4],
        [1, 0, 2, 1, 3, 2, 4, 3],
    ])
    edge_weight = (torch.rand(edge_index.size(1)) > 0.5).float()

    path = osp.join(tmp_path, 'graph.pdf')
    visualize_graph(edge_index, edge_weight, path, backend)
    assert osp.exists(path)


@onlyGraphviz
@pytest.mark.parametrize('backend', [None, 'graphviz'])
def test_visualize_graph_via_graphviz_with_node_labels(tmp_path, backend):
    edge_index = torch.tensor([
        [0, 1, 1, 2, 2, 3, 3, 4],
        [1, 0, 2, 1, 3, 2, 4, 3],
    ])
    edge_weight = (torch.rand(edge_index.size(1)) > 0.5).float()
    node_labels = ['A', 'B', 'C', 'D', 'E']

    path = osp.join(tmp_path, 'graph.pdf')
    visualize_graph(edge_index, edge_weight, path, backend, node_labels)
    assert osp.exists(path)


@withPackage('networkx', 'matplotlib')
@pytest.mark.parametrize('backend', [None, 'networkx'])
def test_visualize_graph_via_networkx(tmp_path, backend):
    edge_index = torch.tensor([
        [0, 1, 1, 2, 2, 3, 3, 4],
        [1, 0, 2, 1, 3, 2, 4, 3],
    ])
    edge_weight = (torch.rand(edge_index.size(1)) > 0.5).float()

    path = osp.join(tmp_path, 'graph.pdf')
    visualize_graph(edge_index, edge_weight, path, backend)
    assert osp.exists(path)



import dgl.function as fn
import torch
import torch.nn.functional as F
from dgl.nn.pytorch import EdgeSoftmax
from torch.nn import Parameter

from torch_geometric.nn.inits import glorot, zeros


class GATConv(torch.nn.Module):
    def __init__(self, g, in_channels, out_channels, heads=1,
                 negative_slope=0.2, dropout=0):
        super().__init__()

        self.g = g
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.heads = heads
        self.negative_slope = negative_slope
        self.dropout = dropout

        self.weight = Parameter(torch.empty(in_channels, heads * out_channels))
        self.att = Parameter(torch.empty(1, heads, 2 * out_channels))
        self.bias = Parameter(torch.empty(heads * out_channels))
        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.weight)
        glorot(self.att)
        zeros(self.bias)

    def gat_msg(self, edge):
        alpha = torch.cat([edge.src['x'], edge.dst['x']], dim=-1)
        alpha = (alpha * self.att).sum(dim=-1)
        alpha = F.leaky_relu(alpha, self.negative_slope)
        return {'m': edge.src['x'], 'a': alpha}

    def gat_reduce(self, node):
        alpha = torch.softmax(node.mailbox['a'], dim=1)
        alpha = F.dropout(alpha, p=self.dropout, training=self.training)
        x = (node.mailbox['m'] * alpha.unsqueeze(-1)).sum(dim=1)
        return {'x': x}

    def forward(self, x):
        x = torch.mm(x, self.weight).view(-1, self.heads, self.out_channels)
        self.g.ndata['x'] = x
        self.g.update_all(self.gat_msg, self.gat_reduce)
        x = self.g.ndata.pop('x')
        x = x.view(-1, self.heads * self.out_channels)
        x = x + self.bias
        return x


class GAT(torch.nn.Module):
    def __init__(self, g, in_channels, out_channels):
        super().__init__()
        self.g = g
        self.conv1 = GATConv(g, in_channels, 8, 8, 0.6, 0.2)
        self.conv2 = GATConv(g, 64, out_channels, 1, 0.6, 0.2)

    def forward(self, x):
        x = F.dropout(x, p=0.6, training=self.training)
        x = F.elu(self.conv1(x))
        x = F.dropout(x, p=0.6, training=self.training)
        x = self.conv2(x)
        return F.log_softmax(x, dim=1)


class GATSPMVConv(torch.nn.Module):
    def __init__(self, g, in_channels, out_channels, heads=1,
                 negative_slope=0.2, dropout=0):
        super().__init__()
        self.g = g
        self.out_channels = out_channels
        self.heads = heads
        self.negative_slope = negative_slope
        self.dropout = dropout
        self.weight = Parameter(torch.empty(in_channels, heads * out_channels))
        self.att_l = Parameter(torch.empty(heads, out_channels, 1))
        self.att_r = Parameter(torch.empty(heads, out_channels, 1))
        self.bias = Parameter(torch.empty(heads * out_channels))
        self.softmax = EdgeSoftmax()
        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.weight)
        glorot(self.att_l)
        glorot(self.att_r)
        zeros(self.bias)

    def forward(self, x):
        x = torch.matmul(x, self.weight)
        x = x.reshape((x.size(0), self.heads, -1))  # NxHxD'
        head_x = x.transpose(0, 1)  # HxNxD'
        a1 = torch.bmm(head_x, self.att_l).transpose(0, 1)  # NxHx1
        a2 = torch.bmm(head_x, self.att_r).transpose(0, 1)  # NxHx1
        self.g.ndata.update({'x': x, 'a1': a1, 'a2': a2})
        self.g.apply_edges(self.edge_attention)
        self.edge_softmax()
        self.g.update_all(fn.src_mul_edge('x', 'a', 'x'), fn.sum('x', 'x'))
        x = self.g.ndata['x'] / self.g.ndata['z']  # NxHxD'
        return x.view(-1, self.heads * self.out_channels)

    def edge_attention(self, edge):
        a = F.leaky_relu(edge.src['a1'] + edge.dst['a2'], self.negative_slope)
        return {'a': a}

    def edge_softmax(self):
        alpha, normalizer = self.softmax(self.g.edata['a'], self.g)
        self.g.ndata['z'] = normalizer
        if self.training and self.dropout > 0:
            alpha = F.dropout(alpha, p=self.dropout, training=True)
        self.g.edata['a'] = alpha


class GATSPMV(torch.nn.Module):
    def __init__(self, g, in_channels, out_channels):
        super().__init__()
        self.g = g
        self.conv1 = GATSPMVConv(g, in_channels, 8, 8, 0.6, 0.2)
        self.conv2 = GATSPMVConv(g, 64, out_channels, 1, 0.6, 0.2)

    def forward(self, x):
        x = F.dropout(x, p=0.6, training=self.training)
        x = F.elu(self.conv1(x))
        x = F.dropout(x, p=0.6, training=self.training)
        x = self.conv2(x)
        return F.log_softmax(x, dim=1)



import time

import torch
import torch.nn.functional as F


def train_runtime(model, data, epochs, device):
    optimizer = torch.optim.Adam(model.parameters(), lr=0.01)
    model = model.to(device)
    data = data.to(device)
    model.train()
    mask = data.train_mask if 'train_mask' in data else data.train_idx
    y = data.y[mask] if 'train_mask' in data else data.train_y

    if torch.cuda.is_available():
        torch.cuda.synchronize()
    elif hasattr(torch.backends, 'mps') and torch.backends.mps.is_available():
        torch.mps.synchronize()
    t_start = time.perf_counter()

    for _ in range(epochs):
        optimizer.zero_grad()
        out = model(data)
        loss = F.nll_loss(out[mask], y)
        loss.backward()
        optimizer.step()

    if torch.cuda.is_available():
        torch.cuda.synchronize()
    elif hasattr(torch.backends, 'mps') and torch.backends.mps.is_available():
        torch.mps.synchronize()
    t_end = time.perf_counter()

    return t_end - t_start



from torch_geometric.datasets.motif_generator import GridMotif


def test_grid_motif():
    motif_generator = GridMotif()
    assert str(motif_generator) == 'GridMotif()'

    motif = motif_generator()
    assert len(motif) == 3
    assert motif.num_nodes == 9
    assert motif.num_edges == 24
    assert motif.edge_index.size() == (2, 24)
    assert motif.edge_index.min() == 0
    assert motif.edge_index.max() == 8
    assert motif.y.size() == (9, )
    assert motif.y.min() == 0
    assert motif.y.max() == 2



import logging
import os

import custom_graphgym  # noqa, register custom modules
import torch

from torch_geometric import seed_everything
from torch_geometric.graphgym.cmd_args import parse_args
from torch_geometric.graphgym.config import (
    cfg,
    dump_cfg,
    load_cfg,
    set_out_dir,
    set_run_dir,
)
from torch_geometric.graphgym.logger import set_printing
from torch_geometric.graphgym.model_builder import create_model
from torch_geometric.graphgym.train import GraphGymDataModule, train
from torch_geometric.graphgym.utils.agg_runs import agg_runs
from torch_geometric.graphgym.utils.comp_budget import params_count
from torch_geometric.graphgym.utils.device import auto_select_device

if __name__ == '__main__':
    # Load cmd line args
    args = parse_args()
    # Load config file
    load_cfg(cfg, args)
    set_out_dir(cfg.out_dir, args.cfg_file)
    # Set Pytorch environment
    torch.set_num_threads(cfg.num_threads)
    dump_cfg(cfg)
    # Repeat for different random seeds
    for _ in range(args.repeat):
        set_run_dir(cfg.out_dir)
        set_printing()
        # Set configurations for each run
        cfg.seed = cfg.seed + 1
        seed_everything(cfg.seed)
        auto_select_device()
        # Set machine learning pipeline
        datamodule = GraphGymDataModule()
        model = create_model()
        # Print model info
        logging.info(model)
        logging.info(cfg)
        cfg.params = params_count(model)
        logging.info('Num parameters: %s', cfg.params)
        train(model, datamodule, logger=True)

    # Aggregate results from different seeds
    agg_runs(cfg.out_dir, cfg.metric_best)
    # When being launched in batch mode, mark a yaml as done
    if args.mark_done:
        os.rename(args.cfg_file, f'{args.cfg_file}_done')



import pytest
import torch

from torch_geometric.profile import benchmark
from torch_geometric.testing import withDevice, withPackage
from torch_geometric.utils.map import map_index


@withDevice
@withPackage('pandas')
@pytest.mark.parametrize('max_index', [3, 100_000_000])
def test_map_index(device, max_index):
    src = torch.tensor([2, 0, 1, 0, max_index], device=device)
    index = torch.tensor([max_index, 2, 0, 1], device=device)

    out, mask = map_index(src, index, inclusive=True)
    assert out.device == device
    assert mask is None
    assert out.tolist() == [1, 2, 3, 2, 0]


@withDevice
@withPackage('pandas')
@pytest.mark.parametrize('max_index', [3, 100_000_000])
def test_map_index_na(device, max_index):
    src = torch.tensor([2, 0, 1, 0, max_index], device=device)
    index = torch.tensor([max_index, 2, 0], device=device)

    out, mask = map_index(src, index, inclusive=False)
    assert out.device == device
    assert mask.device == device
    assert out.tolist() == [1, 2, 2, 0]
    assert mask.tolist() == [True, True, False, True, True]


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--device', type=str, default='cuda')
    args = parser.parse_args()

    src = torch.randint(0, 100_000_000, (100_000, ), device=args.device)
    index = src.unique()

    def trivial_map(src, index, max_index, inclusive):
        if max_index is None:
            max_index = max(src.max(), index.max())

        if inclusive:
            assoc = src.new_empty(max_index + 1)
        else:
            assoc = src.new_full((max_index + 1, ), -1)
        assoc[index] = torch.arange(index.numel(), device=index.device)
        out = assoc[src]

        if inclusive:
            return out, None
        else:
            mask = out != -1
            return out[mask], mask

    print('Inclusive:')
    benchmark(
        funcs=[trivial_map, map_index],
        func_names=['trivial', 'map_index'],
        args=(src, index, None, True),
        num_steps=100,
        num_warmups=50,
    )

    print('Exclusive:')
    benchmark(
        funcs=[trivial_map, map_index],
        func_names=['trivial', 'map_index'],
        args=(src, index[:50_000], None, False),
        num_steps=100,
        num_warmups=50,
    )



import argparse
import asyncio
import hashlib
import json
import logging
import os
import re
import traceback
from copy import deepcopy
from datetime import datetime
from glob import glob
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

# Set environment variables for vLLM multiprocessing compatibility
# Must be set before vLLM imports
os.environ.setdefault('VLLM_WORKER_MULTIPROC_METHOD', 'spawn')
os.environ.setdefault('RAY_EXPERIMENTAL_NOSET_CUDA_VISIBLE_DEVICES', '1')
os.environ.setdefault('RAY_OBJECT_STORE_ALLOW_SLOW_STORAGE', '1')

# Set multiprocessing start method before any other imports
import multiprocessing  # noqa: E402

try:
    multiprocessing.set_start_method('spawn', force=True)
except RuntimeError:
    pass

import yaml  # noqa: E402
from langgraph.graph import StateGraph  # noqa: E402
from langgraph.graph.state import CompiledStateGraph  # noqa: E402

# yapf: disable
from torch_geometric.llm.models.qa_gen import (  # noqa: E402  # isort: skip
    LLMClient,
    QAGenerationState,
    TaskStatus,
    compute_metrics,
    convert_posixpath_to_string,
    create_artifacts,
    end,
    evaluate_qa_pairs,
    feedback_controller,
    filter_by_quality,
    generate_negative_answers_node,
    generate_qa_pairs,
    get_text_artifacts,
    load_input,
    mine_hard_negative_segments,
    monitor_progress,
    route_after_validation,
    should_retry,
    store_results,
    validate_qa_pairs,
)
# yapf: enable

logger = logging.getLogger(__name__)


class QAGenerator:
    def __init__(
        self,
        config_path: str,
        overwrite: bool = False,
        batch: int = 0,
        batch_size: int = 10,
        backend: str = 'nim',
        gen_model: str = 'nvdev/meta/llama-3.1-70b-instruct',
        embedding_model: str = 'nvdev/nvidia/llama-3.2-nv-embedqa-1b-v2',
        evaluation_model: str = 'nvdev/meta/llama-3.1-70b-instruct',
        base_url: str = None,
        api_key: str = None,
        vllm_tensor_parallel_size: int = 1,
        vllm_gpu_memory_utilization: float = 0.9,
        vllm_max_model_len: Optional[int] = None,
        vllm_max_num_batched_tokens: Optional[int] = None,
        model_progression: List[str] = None,
    ):
        self.config = self._load_config(config_path)
        self._validate_data_source_config()
        self.workflow = self._build_qa_generation_workflow()
        self.active_tasks = {}
        self.overwrite = overwrite
        self.batch = batch
        self.batch_size = batch_size
        self.backend = (backend if not self.config.get('backend', None) else
                        self.config.get('backend'))
        self.gen_model = (gen_model if not self.config.get('gen_model', None)
                          else self.config.get('gen_model'))
        self.embedding_model = (embedding_model
                                if not self.config.get('embedding_model', None)
                                else self.config.get('embedding_model'))

        evaluation_obj = self.config.get('evaluation', {})
        self.evaluation_model = (evaluation_model
                                 if not evaluation_obj.get('model', None) else
                                 evaluation_obj.get('model'))

        model_progression = self.config.get('model_progression', None)
        if model_progression is None:
            model_progression = [
                'nvdev/mistralai/mixtral-8x22b-instruct-v0.1',
                'nvdev/meta/llama-3.1-70b-instruct',
                'nvdev/meta/llama-3.1-405b-instruct',
                'nvdev/nvidia/llama-3.1-nemotron-ultra-253b-v1',
            ]
        self.model_progression = model_progression
        self.base_url = base_url

        # Read vLLM config from file if not provided as argument
        vllm_max_model_len = (vllm_max_model_len if vllm_max_model_len
                              is not None else self.config.get(
                                  'vllm_max_model_len', None))
        vllm_max_num_batched_tokens = (
            vllm_max_num_batched_tokens
            if vllm_max_num_batched_tokens is not None else self.config.get(
                'vllm_max_num_batched_tokens', None))

        # Handle gpu_memory_utilization - single value or per-model dict
        gpu_mem_config = self.config.get('vllm_gpu_memory_utilization',
                                         vllm_gpu_memory_utilization)
        if isinstance(gpu_mem_config, dict):
            # Per-model configuration
            vllm_gpu_memory_utilization_generation = gpu_mem_config.get(
                'generation', 0.9)
            vllm_gpu_memory_utilization_embedding = gpu_mem_config.get(
                'embedding', 0.5)
            vllm_gpu_memory_utilization_evaluation = gpu_mem_config.get(
                'evaluation', 0.9)
        else:
            # Single value for all models (backward compatibility)
            vllm_gpu_memory_utilization_generation = gpu_mem_config
            vllm_gpu_memory_utilization_embedding = gpu_mem_config
            vllm_gpu_memory_utilization_evaluation = gpu_mem_config

        # Initialize client based on backend
        gen_util = vllm_gpu_memory_utilization_generation
        emb_util = vllm_gpu_memory_utilization_embedding
        eval_util = vllm_gpu_memory_utilization_evaluation
        self.client = LLMClient(
            generation_model=self.gen_model,
            embedding_model=self.embedding_model,
            evaluation_model=self.evaluation_model,
            backend=self.backend,
            api_key=api_key,
            tensor_parallel_size=vllm_tensor_parallel_size,
            gpu_memory_utilization_generation=gen_util,
            gpu_memory_utilization_embedding=emb_util,
            gpu_memory_utilization_evaluation=eval_util,
            max_model_len=vllm_max_model_len,
            max_num_batched_tokens=vllm_max_num_batched_tokens,
            enable_sleep_mode=True,
        )

    def _build_qa_generation_workflow(self) -> CompiledStateGraph:
        """Build the QA generation workflow using LangGraph."""
        # Create the graph
        workflow = StateGraph(QAGenerationState)

        # Add nodes
        workflow.add_node('load_input', load_input)
        workflow.add_node('create_artifacts', create_artifacts)
        workflow.add_node('generate_qa', generate_qa_pairs)
        workflow.add_node('validate', validate_qa_pairs)
        workflow.add_node('mine_hard_negative_segments',
                          mine_hard_negative_segments)
        workflow.add_node('generate_negative', generate_negative_answers_node)
        workflow.add_node('evaluate_qa',
                          evaluate_qa_pairs)  # New evaluation node
        workflow.add_node('filter_quality', filter_by_quality)
        workflow.add_node('monitor', monitor_progress)
        workflow.add_node('store_results', store_results)
        workflow.add_node('feedback_node', feedback_controller)
        workflow.add_node('end', end)

        # Add edges
        workflow.set_entry_point('load_input')
        workflow.add_edge('load_input', 'create_artifacts')
        workflow.add_edge('create_artifacts', 'generate_qa')
        workflow.add_edge('generate_qa', 'validate')

        # Conditional edges
        workflow.add_conditional_edges(
            'validate',
            route_after_validation,
            {
                'continue': ('mine_hard_negative_segments'
                             ),  # Continue to mine hard negatives
                'feedback': 'feedback_node',  # Route to feedback on failure
            },
        )

        workflow.add_edge(
            'mine_hard_negative_segments',
            'generate_negative')  # Then generate negative answers
        workflow.add_edge('generate_negative',
                          'evaluate_qa')  # Add evaluation after negatives
        workflow.add_edge('evaluate_qa', 'filter_quality')
        workflow.add_edge('filter_quality', 'monitor')
        workflow.add_edge('monitor', 'store_results')

        workflow.add_conditional_edges(
            'store_results',
            lambda x:
            ('feedback'
             if x.get('status') == TaskStatus.FAILED.value else 'end'),
            {
                'feedback_node': 'feedback_node',
                'end': 'end'
            },
        )

        workflow.add_conditional_edges(
            'feedback_node',
            should_retry,
            {
                # "continue": None,
                'feedback_node': 'feedback_node',
            },
        )

        return workflow.compile()

    def _load_config(self, config_path: str) -> Dict:
        with open(config_path) as f:
            return yaml.safe_load(f)

    def _validate_data_source_config(self):
        """Validate that at least one data source is configured."""
        input_dir = self.config.get('input_dir', None)
        huggingface_repo_id = self.config.get('huggingface_repo_id', None)

        assert input_dir or huggingface_repo_id, (
            'Configuration must specify at least one data source: '
            "either 'input_dir' (local path) or 'huggingface_repo_id' "
            '(HuggingFace repository)')

    def _maybe_download_dataset(self) -> Optional[Path]:
        """Check if dataset needs to be downloaded and prompt user.

        Returns:
            Path to input directory if it exists or was downloaded
            successfully, None otherwise.
        """
        input_dir = self.config.get('input_dir', None)

        # If input_dir exists, use it (skip download)
        if input_dir:
            folder = Path(input_dir)
            if folder.exists():
                logger.info('Using existing input directory: %s', input_dir)
                return folder

        # If no HuggingFace config, can't download
        huggingface_repo_id = self.config.get('huggingface_repo_id', None)
        if not huggingface_repo_id:
            return None

        # Prompt user for download
        print(f'Input directory not found: {input_dir}')
        print('Would you like to download the dataset from HuggingFace?')
        print(f'Repository: {huggingface_repo_id}')
        user_input = input('Y/N: ')

        if user_input.lower() in ['y', 'yes']:
            print('Downloading data...')
            downloaded_path = self._download_from_huggingface()
            return downloaded_path
        elif user_input.lower() in ['n', 'no']:
            logger.error('Download declined by user')
            return None
        else:
            logger.error('Invalid user input')
            return None

    def _download_from_huggingface(self) -> Optional[Path]:
        """Download dataset from HuggingFace.

        Returns:
            Path to downloaded directory if successful, None otherwise.
        """
        try:
            import shutil
            import zipfile

            from huggingface_hub import hf_hub_download

            repo_id = self.config.get('huggingface_repo_id')
            repo_type = self.config.get('huggingface_repo_type', 'dataset')
            files_config = self.config.get('huggingface_files', {})
            input_dir = self.config.get('input_dir')

            if not input_dir:
                logger.error(
                    'input_dir must be specified in config for download '
                    'destination')
                return None

            # Create destination directory
            dest_path = Path(input_dir)
            if not dest_path.exists():
                dest_path.mkdir(parents=True, exist_ok=True)

            # Download files based on config
            for file_key, filename in files_config.items():
                logger.info(f'Downloading {file_key}: {filename}')
                downloaded_file = hf_hub_download(
                    repo_id=repo_id,
                    repo_type=repo_type,
                    filename=filename,
                )

                # Handle zip files
                if filename.endswith('.zip'):
                    logger.info(f'Extracting {filename} to {dest_path}')
                    with zipfile.ZipFile(downloaded_file, 'r') as zip_ref:
                        zip_ref.extractall(dest_path)
                else:
                    # Copy non-zip files
                    dest_file = dest_path / filename
                    shutil.copy(downloaded_file, dest_file)
                    logger.info(f'Copied {filename} to {dest_file}')

            logger.info(f'Successfully downloaded dataset to {dest_path}')
            return dest_path

        except Exception as e:
            logger.error(f'Failed to download dataset from HuggingFace: {e}')
            return None

    def check_already_done(self):
        if self.overwrite:
            return []
        output_dir = Path(self.config.get('output_dir'))
        all_qa_files = glob(f'{output_dir}/all_qa_pairs_batch_*.jsonl')
        logger.info('How many all_qa files? %d', len(all_qa_files))
        done_file_paths = []
        for f in all_qa_files:
            with open(f) as fin:
                for line in fin:
                    data = json.loads(line.strip())
                    qa_pairs = data['qa_pairs']
                    for qa_pair in qa_pairs:
                        done_file_paths.append(qa_pair['file_path'])
        done_file_paths = list(set(done_file_paths))
        done_file_paths.sort()
        logger.info('How many files already done? %d', len(done_file_paths))
        return done_file_paths

    async def process_directory(self):
        # Try to download dataset if input_dir doesn't exist
        folder = self._maybe_download_dataset()

        if folder is None or not folder.exists():
            logger.error('Cannot proceed: no input directory available')
            return

        done_file_paths = self.check_already_done()
        output_dir = self.config.get('output_dir')

        input_data = []
        # Use rglob to recursively search in subfolders
        # for common text file extensions
        text_extensions = ['*.txt', '*.text', '*.md']
        for ext in text_extensions:
            input_data.extend(folder.rglob(ext))

        # Also include files without any extension (treat as text files)
        all_files = folder.rglob('*')
        input_data.extend(
            [f for f in all_files if f.is_file() and not f.suffix])

        logger.info('total input files: %d', len(input_data))

        limiter = 5
        batch_size = self.batch_size
        batch = self.batch
        input_data.sort()

        start_idx = batch * batch_size
        end_idx = (batch + 1) * batch_size
        if end_idx > len(input_data):
            end_idx = len(input_data)
        logger.info('input files from %d to %d', start_idx, end_idx)
        input_data = input_data[start_idx:end_idx]

        # Collect results from all chunks
        all_chunk_results = []
        start = 0
        output_dir = Path(self.config.get('output_dir'))
        output_dir.mkdir(exist_ok=True)
        out_file = f'{output_dir}/all_qa_pairs_batch_{batch}.jsonl'
        # Check if file exists before removing
        if os.path.exists(out_file) and self.overwrite:
            os.remove(out_file)

        while True:
            end = (start + limiter if start +
                   limiter < len(input_data) else len(input_data))
            logger.info('Process %d to %d', start, end)
            total_input_data = input_data[start:end]

            input_data_to_do = [
                elem for elem in total_input_data
                if str(elem) not in done_file_paths
            ]

            logger.info(f'Already done: {len(done_file_paths)}, \
                    still_to_do: {len(input_data_to_do)}')

            if len(input_data_to_do) > 0:
                results = await self._process_directory(
                    input_data_to_do, folder)
                # incremental save
                logger.info('results: %d', len(results))
                self.write_all_qa_pairs(results, out_file)
                all_chunk_results.extend(results)
                logger.info('Done from %d to %d', start, end)
            else:
                logger.info('All processed.')
            start = end

            if start >= len(input_data):
                break

        # sanity check
        if not os.path.exists(out_file):
            logger.warning('%s not exist', out_file)
            lines = []
        else:
            with open(out_file) as fin:
                lines = fin.readlines()
        if len(lines) != len(all_chunk_results):
            logger.debug('%s: num lines: %d', out_file, len(lines))
            logger.debug('all_chunk_results: %d', len(all_chunk_results))
            logger.warning(f'{out_file} contains {len(lines)} json objs, \
                    but there are {len(all_chunk_results)} all_chunk_results')

    async def _process_directory(self, input_data_list: List[Any],
                                 input_dir: Optional[Path] = None):
        """Process all transcription files in a directory."""
        # input_data can be a list of Path (filename)
        tasks = []

        self.config.get('output_dir')
        for input_data in input_data_list:
            logger.info('==> input_file: %s', input_data)
            data = input_data
            data_id = input_data
            task = asyncio.create_task(
                self.process_data(data, data_id, input_dir))
            tasks.append(task)

        results = await asyncio.gather(*tasks, return_exceptions=True)

        # Per-chunk summary statistics
        successful = sum(
            1 for r in results
            if isinstance(r, dict) and r.get('status') == TaskStatus.COMPLETED.
            value and r.get('reason') != 'already_processed')
        skipped = sum(
            1 for r in results
            if isinstance(r, dict) and r.get('reason') == 'already_processed')
        failed = sum(1 for r in results if isinstance(r, dict)
                     and r.get('status') == TaskStatus.FAILED.value)

        if self.overwrite:
            logger.info(f'Chunk processing complete (overwrite mode): \
                    {successful} successful, {failed} failed')
        else:
            logger.info(f'Chunk processing complete: {successful} successful, \
                    {skipped} skipped (already processed), {failed} failed')

        return results

    @staticmethod
    def _get_output_stem(file_path: Path, input_dir: Path) -> str:
        """Generate a sanitized output filename stem that \
            includes subfolder paths.

        For files in subfolders, concatenates subfolder names with the \
            filename stem.
        Example: input_dir/subfolder1/subfolder2/file.txt -> \
            subfolder1_subfolder2_file

        Args:
            file_path: Full path to the input file
            input_dir: Base input directory

        Returns:
            Sanitized stem string suitable for use in output filenames
        """
        try:
            # Get the relative path from input_dir to file_path
            relative_path = file_path.relative_to(input_dir)

            # Get all parts (subfolders + filename stem)
            parts = list(relative_path.parent.parts) + [relative_path.stem]

            # Filter out empty parts and join with underscore
            sanitized_parts = [part for part in parts if part and part != '.']
            output_stem = '_'.join(sanitized_parts)

            # Remove any problematic characters for filenames
            output_stem = re.sub(r'[^\w\-_]', '_', output_stem)

            return output_stem
        except ValueError:
            # If file_path is not relative to input_dir, just use the stem
            logger.warning(
                f'File {file_path} is not relative to input_dir {input_dir}, \
                    using stem only')
            return file_path.stem

    async def process_data(self, data: str, data_id: str,
                           input_dir: Optional[Path] = None) -> Dict[str, Any]:
        """Process a single transcription file."""
        # data can be a file_path or an image_b64_str
        # use md5 hash of the file name (without extension) to retrieve the
        # output file
        if os.path.exists(data):
            file_path = data
            task_id = hashlib.md5(Path(file_path).stem.encode()).hexdigest()
            is_image_b64_str = False
        else:
            file_path = data_id
            task_id = data_id
            is_image_b64_str = True

        # Check if output file already exists (resume functionality)
        # Skip this check if --overwrite flag is used
        if not self.overwrite:
            output_dir = Path(self.config.get('output_dir'))
            # Use stem to be consistent with the
            # output file name in store_results
            if os.path.exists(data):
                # Use the new helper to generate output stem with
                # subfolder paths
                if input_dir:
                    stem = self._get_output_stem(Path(file_path), input_dir)
                else:
                    stem = Path(file_path).stem
            else:
                stem = data_id
            output_file = output_dir / f'qa_pairs_{stem}.json'

            if output_file.exists():
                logger.info(f'Skipping {file_path} - output file \
                        already exists: {output_file}')
                return {
                    'task_id': task_id,
                    'status': TaskStatus.COMPLETED.value,
                    'reason': 'already_processed',
                    'file_path': file_path,
                    'is_image_b64_str': is_image_b64_str,
                    'output_file': str(output_file),
                }
        else:
            logger.info(f'Overwrite mode enabled - processing {file_path} \
                    regardless of existing output files')

        try:
            initial_state = {
                'task_id':
                task_id,
                'file_path':
                file_path,
                'input_dir':
                str(input_dir) if input_dir else None,
                'output_dir':
                self.config.get('output_dir'),
                'qa_pairs': [],
                'question_only':
                self.config.get('question_only', False),
                'validation_results': {},
                'negative_answers': [],
                'model_selection':
                '',
                'retry_count':
                0,
                'error_messages': [],
                'status':
                TaskStatus.PENDING.value,
                'metrics': {},
                'feedback': {},
                # "client": self.client,  # Pass OpenAI client
                'num_pairs':
                self.config.get('num_pairs', 10),
                'num_negatives':
                self.config.get('num_negatives', 3),
                'dedup_threshold':
                self.config.get('dedup_threshold', 0.6),
                'parts':
                self.config.get('parts', 3),
                'hard':
                self.config.get('hard', False),
                'use_artifact':
                self.config.get('use_artifact', False),
                'evaluation':
                self.config.get('evaluation'),
                'query_type_distribution':
                self.config.get(
                    'query_type_distribution',
                    '{"multi_hop":0.4,"structural":0.3,"contextual":0.3}',
                ),
                'reasoning_type_distribution':
                self.config.get('reasoning_type_distribution', None),
                'min_hops':
                self.config.get('min_hops', 2),
                'max_hops':
                self.config.get('max_hops', 4),
                'models':
                self.config.get('models'),
                'summary': [],
                'self_contained_question':
                self.config.get('self_contained_question', False),
                'hard_negatives_top_k':
                self.config.get('hard_negatives_top_k', 10),
                'hard_negatives_min_sim':
                self.config.get('hard_negatives_min_sim', 0.5),
                'hard_negatives_max_sim':
                self.config.get('hard_negatives_max_sim', 0.7),
                'enable_hard_negatives':
                self.config.get('enable_hard_negatives', True),
                'text_artifacts':
                None,
            }

            use_artifact = initial_state.get('use_artifact')
            if use_artifact:
                try:
                    ret = await load_input(initial_state)
                    text_artifacts = get_text_artifacts(
                        ret['segments'], self.client)
                    logger.info('text artifacts: %d', len(text_artifacts))
                    initial_state['text_artifacts'] = text_artifacts
                except Exception as e:
                    logger.warning('Failed to retrieve text artifacts: %s', e)

            total_pairs = initial_state['num_pairs']
            pairs_per_shard = int(self.config.get('pairs_per_shard', 0) or 0)
            max_parallel_shards = int(
                self.config.get('max_parallel_shards', 4) or 1)

            # If sharding requested and beneficial, fan out
            # multiple workflow runs
            if pairs_per_shard > 0 and pairs_per_shard < total_pairs:
                logger.info(
                    f'Sharding task {task_id}: total_pairs={total_pairs}, '
                    f'pairs_per_shard={pairs_per_shard}, \
                        max_parallel_shards={max_parallel_shards}')

                shard_states: List[Tuple[Dict[str, Any], Dict[str, Any]]] = []
                remaining = total_pairs
                shard_idx = 0

                while remaining > 0:
                    request_pairs = min(pairs_per_shard, remaining)
                    shard_state = deepcopy(initial_state)
                    shard_state['client'] = self.client
                    shard_task_id = f'{task_id}-shard-{shard_idx}'
                    shard_state['task_id'] = shard_task_id
                    shard_state['num_pairs'] = request_pairs
                    shard_state['parent_task_id'] = task_id
                    shard_state['shard_index'] = shard_idx
                    shard_state['skip_store'] = True
                    shard_states.append((
                        shard_state,
                        {
                            'configurable': {
                                'thread_id': shard_task_id
                            }
                        },
                    ))
                    remaining -= request_pairs
                    shard_idx += 1
                    logger.debug(f'request_pairs: {request_pairs}, \
                            remaining: {remaining}, shard_idx: {shard_idx}')

                semaphore = asyncio.Semaphore(max(1, max_parallel_shards))

                async def run_shard(state_config: Tuple[Dict[str, Any],
                                                        Dict[str, Any]], ):
                    shard_state, shard_config = state_config
                    async with semaphore:
                        logger.info(
                            f'Launching shard {shard_state["task_id"]} '
                            f'for {shard_state["num_pairs"]} pairs')
                        return await self.workflow.ainvoke(
                            shard_state, shard_config)

                shard_tasks = [
                    asyncio.create_task(run_shard(sc)) for sc in shard_states
                ]
                shard_results = await asyncio.gather(*shard_tasks,
                                                     return_exceptions=True)

                successful_results: List[Dict[str, Any]] = []
                for idx, result in enumerate(shard_results):
                    if isinstance(result, Exception):
                        logger.error('Shard %d raised exception: %s', idx,
                                     result)
                        raise result
                    if not isinstance(result, dict):
                        logger.error(
                            f'Shard {idx} returned unexpected result type: \
                                {type(result)}')
                        return {
                            'task_id':
                            task_id,
                            'status':
                            TaskStatus.FAILED.value,
                            'error':
                            (f'Unexpected shard result type: {type(result)}'),
                        }
                    if result.get('status') == TaskStatus.FAILED.value:
                        logger.error(f'Shard {idx} failed with errors: \
                                {result.get("error_messages")}')
                        return result
                    result['client'] = None
                    successful_results.append(result)

                if not successful_results:
                    logger.error('All shards failed for task %s', task_id)
                    return {
                        'task_id': task_id,
                        'status': TaskStatus.FAILED.value,
                        'error': 'All shards failed to produce output',
                    }

                # Merge shard outputs
                aggregate_state = deepcopy(successful_results[0])
                aggregate_state['task_id'] = task_id
                aggregate_state['parent_task_id'] = task_id
                aggregate_state['skip_store'] = False
                aggregate_state['shard_index'] = None
                aggregate_state['shard_results'] = [
                    res.get('task_id') for res in successful_results
                ]
                aggregate_state['enable_hard_negatives'] = initial_state.get(
                    'enable_hard_negatives', True)

                for key in (
                        'qa_pairs',
                        'validated_pairs',
                        'high_quality_qa_pairs',
                        'low_quality_qa_pairs',
                        'not_evaluated_pairs',
                ):
                    aggregate_state[key] = []

                def make_key(pair: Dict[str, Any]) -> Tuple[str, str]:
                    return (
                        pair.get('question', '').strip(),
                        pair.get('answer', '').strip(),
                    )

                merged_pairs: Dict[Tuple[str, str], Dict[str, Any]] = {}
                qa_order: List[Tuple[str, str]] = []
                validated_keys: List[Tuple[str, str]] = []
                high_keys: List[Tuple[str, str]] = []
                low_keys: List[Tuple[str, str]] = []
                not_eval_keys: List[Tuple[str, str]] = []

                for shard_result in successful_results:
                    for pair in shard_result.get('qa_pairs', []) or []:
                        key = make_key(pair)
                        if key not in merged_pairs:
                            merged_pairs[key] = deepcopy(pair)
                        if key not in qa_order:
                            qa_order.append(key)

                    for pair in shard_result.get('validated_pairs', []) or []:
                        key = make_key(pair)
                        if key not in merged_pairs:
                            merged_pairs[key] = deepcopy(pair)
                        if key not in validated_keys:
                            validated_keys.append(key)

                    for pair in (shard_result.get('high_quality_qa_pairs', [])
                                 or []):
                        key = make_key(pair)
                        if key not in merged_pairs:
                            merged_pairs[key] = deepcopy(pair)
                        if key not in high_keys:
                            high_keys.append(key)

                    for pair in (shard_result.get('low_quality_qa_pairs', [])
                                 or []):
                        key = make_key(pair)
                        if key not in merged_pairs:
                            merged_pairs[key] = deepcopy(pair)
                        if key not in low_keys:
                            low_keys.append(key)

                    for pair in (shard_result.get('not_evaluated_pairs', [])
                                 or []):
                        key = make_key(pair)
                        if key not in merged_pairs:
                            merged_pairs[key] = deepcopy(pair)
                        if key not in not_eval_keys:
                            not_eval_keys.append(key)

                ordered_keys: List[Tuple[str, str]] = []

                def append_keys(source: List[Tuple[str, str]]):
                    for key in source:
                        if key in merged_pairs and key not in ordered_keys:
                            ordered_keys.append(key)

                append_keys(high_keys)
                append_keys(not_eval_keys)
                append_keys(low_keys)
                append_keys(qa_order)

                if total_pairs and len(ordered_keys) > total_pairs:
                    ordered_keys = ordered_keys[:total_pairs]

                allowed_keys = set(ordered_keys)

                aggregate_state['qa_pairs'] = [
                    merged_pairs[key] for key in ordered_keys
                ]
                aggregate_state['validated_pairs'] = [
                    merged_pairs[key] for key in validated_keys
                    if key in allowed_keys
                ]
                aggregate_state['high_quality_qa_pairs'] = [
                    merged_pairs[key] for key in high_keys
                    if key in allowed_keys
                ]
                aggregate_state['low_quality_qa_pairs'] = [
                    merged_pairs[key] for key in low_keys
                    if key in allowed_keys
                ]
                aggregate_state['not_evaluated_pairs'] = [
                    merged_pairs[key] for key in not_eval_keys
                    if key in allowed_keys
                ]

                for bucket in (
                        'qa_pairs',
                        'validated_pairs',
                        'high_quality_qa_pairs',
                        'low_quality_qa_pairs',
                        'not_evaluated_pairs',
                ):
                    for pair in aggregate_state[bucket]:
                        pair['task_id'] = task_id

                enable_hard_negatives = aggregate_state.get(
                    'enable_hard_negatives', True)
                total_hard_negs = (sum(
                    len(pair.get('hard_negative_segment_ids', []))
                    for pair in aggregate_state['qa_pairs'])
                                   if enable_hard_negatives else 0)
                avg_hard_negs = (total_hard_negs /
                                 len(aggregate_state['qa_pairs'])
                                 if enable_hard_negatives
                                 and aggregate_state['qa_pairs'] else 0.0)
                aggregate_state['hard_negatives_stats'] = {
                    'total_hard_negatives': total_hard_negs,
                    'avg_per_question': avg_hard_negs,
                    'enabled': enable_hard_negatives,
                }

                # Update validation metrics
                total_validated = sum(
                    res.get('validation_results', {}).get(
                        'validated_pairs', 0) for res in successful_results)
                total_attempted = sum(
                    res.get('validation_results', {}).get('total_pairs', 0)
                    for res in successful_results)
                if total_attempted:
                    aggregate_state['validation_results'] = {
                        'total_pairs':
                        total_attempted,
                        'validated_pairs':
                        total_validated,
                        'dropped_pairs':
                        total_attempted - total_validated,
                        'validation_method': (successful_results[0].get(
                            'validation_results',
                            {}).get('validation_method', 'hybrid_with_faiss')),
                    }

                # Recompute evaluation and quality metrics
                # based on combined lists
                evaluated_pairs = (aggregate_state['high_quality_qa_pairs'] +
                                   aggregate_state['low_quality_qa_pairs'])
                if evaluated_pairs:
                    avg_score = sum(
                        qa.get('evaluation', {}).get('overall', {}).get(
                            'score', 0)
                        for qa in evaluated_pairs) / len(evaluated_pairs)
                else:
                    avg_score = 0.0

                sample_eval = next(
                    (res.get('evaluation_metrics')
                     for res in successful_results
                     if res.get('evaluation_metrics')),
                    {},
                )

                aggregate_state['evaluation_metrics'] = {
                    'average_quality_score':
                    avg_score,
                    'evaluated_pairs':
                    len(evaluated_pairs),
                    'evaluation_model':
                    sample_eval.get('evaluation_model'),
                    'high_quality_ratio':
                    (len(aggregate_state['high_quality_qa_pairs']) /
                     max(len(evaluated_pairs), 1) if evaluated_pairs else 0.0),
                }

                quality_threshold = initial_state.get('evaluation', {}).get(
                    'quality_threshold', 6.0)
                aggregate_state['quality_metrics'] = {
                    'total_pairs':
                    len(aggregate_state['qa_pairs']),
                    'evaluated_pairs':
                    len(evaluated_pairs),
                    'high_quality_pairs':
                    len(aggregate_state['high_quality_qa_pairs']),
                    'low_quality_pairs':
                    len(aggregate_state['low_quality_qa_pairs']),
                    'not_evaluated_pairs':
                    len(aggregate_state['not_evaluated_pairs']),
                    'quality_threshold':
                    quality_threshold,
                    'average_quality':
                    avg_score,
                }

                aggregate_state['metrics'] = {
                    'task_id': task_id,
                    'qa_pairs_generated': len(aggregate_state['qa_pairs']),
                    'shards_completed': len(successful_results),
                    'pairs_per_shard': pairs_per_shard,
                    'total_requested_pairs': total_pairs,
                    'timestamp': datetime.now().isoformat(),
                }

                aggregate_state['qa_data_to_write'] = []
                for shard_result in successful_results:
                    qa_data_to_write = shard_result.get('qa_data_to_write', [])
                    logger.debug(
                        f'shard qa_data_to_write: {len(qa_data_to_write)}')
                    aggregate_state['qa_data_to_write'].extend(
                        qa_data_to_write)
                logger.info(f'total shard qa_data_to_write: \
                        {len(aggregate_state["qa_data_to_write"])}')

                aggregate_state['status'] = TaskStatus.COMPLETED.value

                # Persist combined results once
                store_response = await store_results(aggregate_state)
                aggregate_state.update(store_response or {})
                return aggregate_state

            # Fallback: single workflow run
            config = {'configurable': {'thread_id': task_id}}
            initial_state['client'] = self.client
            final_state = await self.workflow.ainvoke(initial_state, config)

            return final_state
        except Exception as e:
            error_msg = traceback.format_exc()
            logger.error('Error processing file %s: %s, %s', file_path, e,
                         error_msg)
            return {
                'task_id': task_id,
                'status': TaskStatus.FAILED.value,
                'error': str(e),
            }

    def write_all_qa_pairs(self, all_chunk_results, out_file):
        # Aggregate and write all QA pairs after processing all chunks

        def update_dict(parent_dict, key, val):
            if isinstance(val, dict):
                parent_dict.update(val)
            elif isinstance(val, list):
                parent_dict[key] = val

        try:
            combined_results = []
            total_qa_pairs = 0
            for result in all_chunk_results:
                if isinstance(result, dict) and result.get('qa_data_to_write'):
                    combined_result = {}
                    combined_result['qa_pairs'] = result.get(
                        'qa_data_to_write')
                    total_qa_pairs += len(combined_result['qa_pairs'])

                    top_artifacts = result.get('top_artifacts', {})
                    evaluation_metrics = result.get('evaluation_metrics', {})
                    quality_metrics = result.get('quality_metrics', {})
                    metrics = result.get('metrics', {})
                    segments = result.get('segments', [])

                    update_dict(combined_result, 'top_artifacts',
                                top_artifacts)
                    update_dict(
                        combined_result,
                        'evaluation_metrics',
                        evaluation_metrics,
                    )
                    update_dict(combined_result, 'quality_metrics',
                                quality_metrics)
                    update_dict(combined_result, 'metrics', metrics)
                    update_dict(combined_result, 'segments', segments)

                    combined_results.append(combined_result)

            logger.info(f'how many combined_results? {len(combined_results)}, \
                    total_qa_pairs: {total_qa_pairs}')

            if combined_results:
                # Get distributions from config
                query_type_dist = self.config.get(
                    'query_type_distribution',
                    '{"multi_hop":0.4,"structural":0.3,"contextual":0.3}',
                )
                reasoning_type_dist = self.config.get(
                    'reasoning_type_distribution', None)

                question_only = self.config.get('question_only', False)
                logger.debug('question_only? %s', question_only)
                if not question_only:
                    logger.debug(
                        f'combined results[0]: {combined_results[0].keys()}')
                    all_qa_pairs = []
                    for res in combined_results:
                        all_qa_pairs.extend(res['qa_pairs'])
                        logger.info('num all_qa_pairs: %d', len(all_qa_pairs))
                    metrics = compute_metrics(all_qa_pairs, query_type_dist,
                                              reasoning_type_dist)
                    logger.info(f'Final aggregated metrics across all chunks: \
                            {metrics}')
                else:
                    logger.info('Skip compute_metrics')

                # append to jsonl
                with open(out_file, 'a') as fout:
                    for elem in combined_results:
                        line = json.dumps(convert_posixpath_to_string(elem))
                        fout.write(line + '\n')
                    # json.dump(all_qa_pairs, fout, indent=2)
                    logger.info(
                        f'Wrote {len(combined_results)} QA pairs to {out_file}'
                    )

            else:
                logger.warning('No QA pairs generated across all chunks')

        except Exception as e:
            error_msg = traceback.format_exc()
            logger.error('Error write_all_qa_pairs: %s, %s', e, error_msg)

    def visualize_workflow(self):
        """Generate a visual representation of the workflow."""
        try:
            # This would generate a Mermaid diagram or similar
            return self.workflow.get_graph().draw_mermaid()
        except Exception as e:
            logger.error(e)
            return 'Visualization not available'

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit with cleanup."""
        if hasattr(self, 'client') and self.client:
            self.client.cleanup()
        return False


if __name__ == '__main__':
    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description='TrueQuery QA Generation System')
    parser.add_argument(
        '--config',
        '-c',
        type=str,
        required=True,
        default='txt2qa_config/text_config.yaml',
        help='Path to configuration YAML file (required).',
    )
    parser.add_argument(
        '--overwrite',
        action='store_true',
        help='Overwrite existing output files and reprocess all input files \
            (disables resume functionality).',
    )

    parser.add_argument(
        '--batch',
        type=int,
        default=0,
        help='Which batch to run. For example, if batch_size = 10, \
            we will process input files from 0 to 9 with batch 0, \
                and 10 to 19 with batch 1 etc',
    )

    parser.add_argument(
        '--batch-size',
        type=int,
        default=10,
        help='Number of input files to process in each batch',
    )

    parser.add_argument(
        '--verbose',
        '-v',
        action='store_true',
        help='Enable verbose logging (INFO and DEBUG messages).\
            By default, only WARNING/ERROR/CRITICAL are shown.',
    )

    args = parser.parse_args()

    # Configure logging based on verbose flag
    if args.verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        )

    # Load configuration from file
    config_path = args.config
    if not os.path.exists(config_path):
        logger.error('Configuration file not found: %s', config_path)
        exit(1)

    logger.info('Loading configuration from: %s', config_path)

    # Run the system with proper resource cleanup
    with QAGenerator(
            config_path,
            overwrite=args.overwrite,
            batch=args.batch,
            batch_size=args.batch_size,
    ) as generator:
        # Process files
        asyncio.run(generator.process_directory())
        logger.info('Processing complete')
        # logger.info(generator.visualize_workflow())



import pytest
import torch

import torch_geometric.typing
from torch_geometric.nn import GeneralConv
from torch_geometric.typing import SparseTensor


@pytest.mark.parametrize('kwargs', [
    dict(),
    dict(skip_linear=True),
    dict(directed_msg=False),
    dict(heads=3),
    dict(attention=True),
    dict(heads=3, attention=True),
    dict(heads=3, attention=True, attention_type='dot_product'),
    dict(l2_normalize=True),
])
def test_general_conv(kwargs):
    x = torch.randn(4, 8)
    edge_index = torch.tensor([[0, 1, 2, 3], [0, 0, 1, 1]])
    edge_attr = torch.randn(edge_index.size(1), 16)

    conv = GeneralConv(8, 32, **kwargs)
    assert str(conv) == 'GeneralConv(8, 32)'

    out = conv(x, edge_index)
    assert out.size() == (4, 32)

    if torch_geometric.typing.WITH_TORCH_SPARSE:
        adj = SparseTensor.from_edge_index(edge_index, sparse_sizes=(4, 4))
        assert torch.allclose(conv(x, adj.t()), out, atol=1e-6)

    conv = GeneralConv(8, 32, in_edge_channels=16, **kwargs)
    assert str(conv) == 'GeneralConv(8, 32)'

    out = conv(x, edge_index, edge_attr)
    assert out.size() == (4, 32)

    if torch_geometric.typing.WITH_TORCH_SPARSE:
        adj = SparseTensor.from_edge_index(edge_index, edge_attr, (4, 4))
        assert torch.allclose(conv(x, adj.t()), out, atol=1e-6)
